# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2024 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Commands related to purging of old data
"""

import os
import datetime
import shutil
import logging


log = logging.getLogger(__name__)


def run_purge(config, purge_title, purge_title_plural, thing_finder, thing_purger,
              before=None, before_days=None, default_before_days=90,
              dry_run=False, progress=None):
    from rattail.db.util import finalize_session

    log.info("will purge things of type: %s", purge_title)

    if before and before_days:
        log.warning("specifying both --before and --before-days is "
                    "redundant; --before will take precedence.")

    app = config.get_app()
    session = app.make_session()

    # calculate our cutoff date
    if before:
        cutoff = before
    else:
        today = app.today()
        cutoff = today - datetime.timedelta(days=before_days or default_before_days)
    cutoff = datetime.datetime.combine(cutoff, datetime.time(0))
    cutoff = app.localtime(cutoff)
    log.info("using %s as cutoff date", cutoff.date())

    # find things, and purge them
    things = thing_finder(session, cutoff, dry_run=dry_run)
    log.info("found %s thing(s) to purge", len(things or []))
    if things:
        purged = purge_things(config, session, things, thing_purger, cutoff, purge_title_plural,
                              dry_run=dry_run, progress=progress)
        log.info("%spurged %s %s",
                 "(would have) " if dry_run else "",
                 purged, purge_title_plural)

    finalize_session(session, dry_run=dry_run)


def purge_things(config, session, things, purger, cutoff, purge_title_plural,
                 dry_run=False, progress=None):
    app = config.get_app()
    result = app.make_object(purged=0)

    def purge(thing, i):
        if purger(session, thing, cutoff, dry_run=dry_run):
            result.purged += 1
        if i % 200 == 0:
            session.flush()

    app.progress_loop(purge, things, progress,
                      message=f"Purging {purge_title_plural}")
    return result.purged



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2024 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Core Stuff
"""

import warnings

from wuttjamaican.util import make_uuid


# Arbitrary object used to explicitly denote an unspecified value, e.g. for
# comparison to an argument within a function signature.
UNSPECIFIED = object()


class Object(object):
    """
    Generic base class.

    This is used mostly for convenience.
    """

    def __init__(self, **kwargs):
        """
        Constructor which assigns keyword arguments as attributes.
        """
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __getitem__(self, key):
        """
        Allows dict-like access to the object's attributes.
        """
        if hasattr(self, key):
            return getattr(self, key)

    def setdefault(self, key, value):
        if not hasattr(self, key):
            setattr(self, key, value)


def get_uuid():
    """
    DEPRECATED; use
    :func:`~wuttjamaican:wuttjamaican.util.make_uuid()` instead.
    """
    warnings.warn("rattail.core.get_uuid() is deprecated; "
                  "please use wuttjamaican.util.make_uuid() instead",
                  DeprecationWarning, stacklevel=2)
    return make_uuid()



# -*- coding: utf-8; -*-

from unittest import TestCase

import pytest

from rattail.config import make_config
from rattail.autocomplete import base as mod
from rattail.db import Session


class TestAutocompleter(TestCase):

    def setUp(self):
        self.config = self.make_config()
        self.app = self.config.get_app()

    def make_config(self):
        return make_config([], extend=False)

    def test_constructor(self):

        # cannot instantiate autocompleter with no key
        class BadAutocompleter(mod.Autocompleter):
            pass
        self.assertRaises(NotImplementedError, BadAutocompleter, self.config)

        # but works okay if key is defined
        class GoodAutocompleter(mod.Autocompleter):
            autocompleter_key = 'stores'
        autocompleter = GoodAutocompleter(self.config)
        self.assertIsNotNone(autocompleter)

    def test_get_model_class(self):
        try:
            model = self.app.model
        except ImportError:
            pytest.skip("test is not relevant without sqlalchemy")

        # no model class by default; hence error
        class BadAutocompleter(mod.Autocompleter):
            autocompleter_key = 'stores'
        autocompleter = BadAutocompleter(self.config)
        self.assertRaises(AttributeError, autocompleter.get_model_class)

        # but if one is set, it is returned
        class GoodAutocompleter(mod.Autocompleter):
            autocompleter_key = 'stores'
            model_class = model.Store
        autocompleter = GoodAutocompleter(self.config)
        result = autocompleter.get_model_class()
        self.assertIs(result, model.Store)

    def test_autocomplete_fieldname(self):

        # no fieldname by default; hence error
        class BadAutocompleter(mod.Autocompleter):
            autocompleter_key = 'stores'
        autocompleter = BadAutocompleter(self.config)
        self.assertRaises(NotImplementedError, getattr,
                          autocompleter, 'autocomplete_fieldname')

        # but works okay if class defines it
        class GoodAutocompleter(mod.Autocompleter):
            autocompleter_key = 'stores'
            autocomplete_fieldname = 'name'
        autocompleter = GoodAutocompleter(self.config)
        self.assertEqual(autocompleter.autocomplete_fieldname, 'name')

    def test_autocomplete(self):
        try:
            import sqlalchemy as sa
        except ImportError:
            pytest.skip("test is not relevant without sqlalchemy")

        engine = sa.create_engine('sqlite://')
        model = self.app.model
        model.Base.metadata.create_all(bind=engine)
        session = Session(bind=engine)

        class StoreAutocompleter(mod.Autocompleter):
            autocompleter_key = 'store'
            model_class = model.Store
            autocomplete_fieldname = 'name'

        # first create a few stores
        store001 = model.Store(id='001', name="Acme Store #1")
        session.add(store001)
        store002 = model.Store(id='002', name="Acme Store #2")
        session.add(store002)
        store999 = model.Store(id='999', name="Warehouse")
        session.add(store999)

        # and our autocompleter
        autocompleter = StoreAutocompleter(self.config)

        # searching for nothing yields no results
        result = autocompleter.autocomplete(session, '')
        self.assertEqual(len(result), 0)

        # search for 'acme' yields stores 1, 2
        result = autocompleter.autocomplete(session, 'acme')
        self.assertEqual(len(result), 2)
        uuids = [info['value'] for info in result]
        self.assertIn(store001.uuid, uuids)
        self.assertIn(store002.uuid, uuids)

        # search for 'warehouse' yields store 999
        result = autocompleter.autocomplete(session, 'warehouse')
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0]['value'], store999.uuid)

        # search for 'bogus' yields no results
        result = autocompleter.autocomplete(session, 'bogus')
        self.assertEqual(len(result), 0)

        # search for 's' yields all 3 stores
        result = autocompleter.autocomplete(session, 's')
        self.assertEqual(len(result), 3)

        # unless we cap the max results
        autocompleter.max_results = 2
        result = autocompleter.autocomplete(session, 's')
        self.assertEqual(len(result), 2)

        # removing cap should get all 3 again
        autocompleter.max_results = None
        result = autocompleter.autocomplete(session, 's')
        self.assertEqual(len(result), 3)


class TestPhoneMagicMixin(TestCase):

    def setUp(self):
        self.config = self.make_config()
        self.app = self.config.get_app()

    def make_config(self):
        return make_config([], extend=False)

    def test_autocomplete(self):
        try:
            import sqlalchemy as sa
        except ImportError:
            pytest.skip("test is not relevant without sqlalchemy")

        engine = sa.create_engine('sqlite://')
        model = self.app.model
        model.Base.metadata.create_all(bind=engine)
        session = Session(bind=engine)

        class StoreMagicAutocompleter(mod.PhoneMagicMixin, mod.Autocompleter):
            autocompleter_key = 'store'
            model_class = model.Store
            autocomplete_fieldname = 'name'
            phone_model_class = model.StorePhoneNumber

        # first create a few stores, with phones
        store001 = model.Store(id='001', name="Acme Store #1")
        store001.add_phone_number('417-555-0001')
        session.add(store001)
        store002 = model.Store(id='002', name="Acme Store #2")
        store002.add_phone_number('417-555-0002')
        session.add(store002)
        store999 = model.Store(id='999', name="Warehouse")
        session.add(store999)

        # and our autocompleter
        autocompleter = StoreMagicAutocompleter(self.config)

        # searching for nothing yields no results
        result = autocompleter.autocomplete(session, '')
        self.assertEqual(len(result), 0)

        # search for 'acme' yields stores 1, 2
        result = autocompleter.autocomplete(session, 'acme')
        self.assertEqual(len(result), 2)
        uuids = [info['value'] for info in result]
        self.assertIn(store001.uuid, uuids)
        self.assertIn(store002.uuid, uuids)

        # search for 'warehouse' yields store 999
        result = autocompleter.autocomplete(session, 'warehouse')
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0]['value'], store999.uuid)

        # TODO: need to test the actual phone number magic, but sqlite
        # does not have a regexp_replace() function built-in.  so
        # either need to code some workaround, or require the 're'
        # extension be loaded for sqlite, etc.



# -*- coding: utf-8; -*-

from unittest import TestCase

try:
    from rattail.db import config as conf
except ImportError:
    pass
else:

    class TestMakeEngineFromConfig(TestCase):

        def test_record_changes(self):

            # no attribute is set by default
            engine = conf.make_engine_from_config({
                'sqlalchemy.url': 'sqlite://',
            })
            self.assertRaises(AttributeError, getattr, engine, 'rattail_record_changes')

            # but if flag is true, attr is set
            engine = conf.make_engine_from_config({
                'sqlalchemy.url': 'sqlite://',
                'sqlalchemy.record_changes': 'true'
            })
            self.assertTrue(engine.rattail_record_changes)

        def test_log_pool_status(self):

            # no attribute is set by default
            engine = conf.make_engine_from_config({
                'sqlalchemy.url': 'sqlite://',
            })
            self.assertRaises(AttributeError, getattr, engine, 'rattail_log_pool_status')

            # but if flag is true, attr is set
            engine = conf.make_engine_from_config({
                'sqlalchemy.url': 'sqlite://',
                'sqlalchemy.log_pool_status': 'true'
            })
            self.assertTrue(engine.rattail_log_pool_status)



# -*- coding: utf-8; -*-

from unittest import TestCase


from rattail.config import make_config

try:
    import sqlalchemy as sa
    from rattail.autocomplete import employees as mod
    from rattail.db import Session
except ImportError:
    pass
else:

    class TestEmployeeAutocompleter(TestCase):

        def setUp(self):
            self.config = self.make_config()
            self.app = self.config.get_app()
            self.autocompleter = self.make_autocompleter()

        def make_config(self):
            return make_config([], extend=False)

        def make_autocompleter(self):
            return mod.EmployeeAutocompleter(self.config)

        def test_autocomplete(self):
            engine = sa.create_engine('sqlite://')
            model = self.app.model
            model.Base.metadata.create_all(bind=engine)
            session = Session(bind=engine)
            enum = self.config.get_enum()

            # first create some employees
            alice = model.Person(display_name='Alice Chalmers')
            alice.employee = model.Employee(status=enum.EMPLOYEE_STATUS_CURRENT)
            session.add(alice)
            bob = model.Person(display_name='Bob Loblaw')
            bob.employee = model.Employee(status=enum.EMPLOYEE_STATUS_CURRENT)
            session.add(bob)
            charlie = model.Person(display_name='Charlie Chaplin')
            charlie.employee = model.Employee(status=enum.EMPLOYEE_STATUS_FORMER)
            session.add(charlie)

            # searching for nothing yields no results
            result = self.autocompleter.autocomplete(session, '')
            self.assertEqual(len(result), 0)

            # search for 'l' yields only 2 current employees
            result = self.autocompleter.autocomplete(session, 'l')
            self.assertEqual(len(result), 2)

            # search for 'alice' yields just Alice Chalmers
            result = self.autocompleter.autocomplete(session, 'alice')
            self.assertEqual(len(result), 1)
            self.assertEqual(result[0]['value'], alice.employee.uuid)



# -*- coding: utf-8; -*-

import os
from unittest import TestCase

from rattail.batch import handlers as mod
from rattail.config import make_config
from rattail.db import Session

try:
    import sqlalchemy as sa
except ImportError:
    pass
else:

    class TestBatchHandler(TestCase):

        def setUp(self):
            self.config = self.make_config()
            self.app = self.config.get_app()
            self.handler = self.make_handler()

        def make_config(self):
            return make_config([], extend=False)

        def make_handler(self):
            return mod.BatchHandler(self.config)

        def test_consume_batch_id(self):
            engine = sa.create_engine('sqlite://')
            model = self.app.model
            model.Base.metadata.create_all(bind=engine)
            session = Session(bind=engine)

            # first id is 1
            result = self.handler.consume_batch_id(session)
            self.assertEqual(result, 1)

            # second is 2; test string version
            result = self.handler.consume_batch_id(session, as_str=True)
            self.assertEqual(result, '00000002')

        def test_get_effective_rows(self):
            engine = sa.create_engine('sqlite://')
            model = self.app.model
            model.Base.metadata.create_all(bind=engine)
            session = Session(bind=engine)

            # make batch w/ 3 rows
            user = model.User(username='patty')
            batch = model.NewProductBatch(id=1, created_by=user)
            batch.data_rows.append(model.NewProductBatchRow())
            batch.data_rows.append(model.NewProductBatchRow())
            batch.data_rows.append(model.NewProductBatchRow())
            self.assertEqual(len(batch.data_rows), 3)

            # all rows should be effective by default
            result = self.handler.get_effective_rows(batch)
            self.assertEqual(len(result), 3)

            # unless we mark one as "removed"
            batch.data_rows[1].removed = True
            result = self.handler.get_effective_rows(batch)
            self.assertEqual(len(result), 2)

            # or if we delete one
            batch.data_rows.pop(-1)
            result = self.handler.get_effective_rows(batch)
            self.assertEqual(len(result), 1)



# -*- coding: utf-8; -*-

from unittest import TestCase
from unittest.mock import patch, Mock

from rattail.config import make_config

try:
    import sqlalchemy as sa
    from sqlalchemy import orm
    from sqlalchemy.orm.exc import MultipleResultsFound
    from rattail.importing import sqlalchemy as saimport
except ImportError:
    pass
else:

    class Widget(object):
        pass

    metadata = sa.MetaData()

    widget_table = sa.Table('widgets', metadata,
                            sa.Column('id', sa.Integer(), primary_key=True),
                            sa.Column('description', sa.String(length=50)))

    widget_mapper = orm.mapper(Widget, widget_table)

    WIDGETS = [
        {'id': 1, 'description': "Main Widget"},
        {'id': 2, 'description': "Other Widget"},
        {'id': 3, 'description': "Other Widget"},
    ]


    class TestFromSQLAlchemy(TestCase):

        def setUp(self):
            self.config = self.make_config()

        def make_config(self):
            config = make_config([], extend=False)
            config.setdefault('rattail', 'timezone.default', 'America/Chicago')
            return config

        def test_query(self):
            Session = orm.sessionmaker(bind=sa.create_engine('sqlite://'))
            session = Session()
            importer = saimport.FromSQLAlchemy(self.config,
                                               host_session=session,
                                               host_model_class=Widget)
            self.assertEqual(str(importer.query()),
                             "SELECT widgets.id AS widgets_id, widgets.description AS widgets_description \n"
                             "FROM widgets")


    class TestToSQLAlchemy(TestCase):

        def setUp(self):
            self.config = self.make_config()
            engine = sa.create_engine('sqlite://')
            metadata.create_all(bind=engine)
            Session = orm.sessionmaker(bind=engine)
            self.session = Session()
            for data in WIDGETS:
                widget = Widget()
                for key, value in data.items():
                    setattr(widget, key, value)
                self.session.add(widget)
            self.session.commit()

        def tearDown(self):
            self.session.close()
            del self.session

        def make_config(self):
            config = make_config([], extend=False)
            config.setdefault('rattail', 'timezone.default', 'America/Chicago')
            return config

        def make_importer(self, **kwargs):
            kwargs.setdefault('config', self.config)
            kwargs.setdefault('model_class', Widget)
            return saimport.ToSQLAlchemy(**kwargs)

        def test_model_mapper(self):
            importer = self.make_importer()
            self.assertIs(importer.model_mapper, widget_mapper)

        def test_model_table(self):
            importer = self.make_importer()
            self.assertIs(importer.model_table, widget_table)

        def test_simple_fields(self):
            importer = self.make_importer()
            self.assertEqual(importer.simple_fields, ['id', 'description'])

        def test_get_single_local_object(self):

            # simple key
            importer = self.make_importer(key='id', session=self.session)
            widget = importer.get_single_local_object((1,))
            self.assertEqual(widget.id, 1)
            self.assertEqual(widget.description, "Main Widget")

            # compound key
            importer = self.make_importer(key=('id', 'description'), session=self.session)
            widget = importer.get_single_local_object((1, "Main Widget"))
            self.assertEqual(widget.id, 1)
            self.assertEqual(widget.description, "Main Widget")

            # widget not found
            importer = self.make_importer(key='id', session=self.session)
            self.assertIsNone(importer.get_single_local_object((42,)))

            # multiple widgets found
            importer = self.make_importer(key='description', session=self.session)
            self.assertRaises(MultipleResultsFound, importer.get_single_local_object, ("Other Widget",))

        def test_create_object(self):
            importer = self.make_importer(key='id', session=self.session)
            widget = importer.create_object((42,), {'id': 42, 'description': "Latest Widget"})
            self.assertIn(widget, self.session)
            self.assertIn(widget, self.session.new)
            self.assertEqual(widget.id, 42)
            self.assertEqual(widget.description, "Latest Widget")

        def test_delete_object(self):
            widget = self.session.query(Widget).get(1)
            self.assertIn(widget, self.session)
            importer = self.make_importer(session=self.session)
            self.assertTrue(importer.delete_object(widget))
            self.assertIn(widget, self.session)
            self.assertIn(widget, self.session.deleted)
            self.session.flush()
            self.assertIsNone(self.session.query(Widget).get(1))

        def test_cache_model(self):
            importer = self.make_importer(key='id', session=self.session)
            cache = importer.cache_model(Widget, key='id')
            self.assertEqual(len(cache), 3)
            for i in range(1, 4):
                self.assertIn(i, cache)
                self.assertIsInstance(cache[i], Widget)
                self.assertEqual(cache[i].id, i)
                self.assertEqual(cache[i].description, WIDGETS[i-1]['description'])

        def test_cache_local_data(self):
            importer = self.make_importer(key='id', session=self.session)
            importer.handler = Mock(local_title="Nevermind")
            cache = importer.cache_local_data()
            self.assertEqual(len(cache), 3)
            for i in range(1, 4):
                self.assertIn((i,), cache)
                cached = cache[(i,)]
                self.assertIsInstance(cached, dict)
                self.assertIsInstance(cached['object'], Widget)
                self.assertEqual(cached['object'].id, i)
                self.assertEqual(cached['object'].description, WIDGETS[i-1]['description'])
                self.assertIsInstance(cached['data'], dict)
                self.assertEqual(cached['data']['id'], i)
                self.assertEqual(cached['data']['description'], WIDGETS[i-1]['description'])

        # TODO: lame
        def test_flush_session(self):
            importer = self.make_importer(fields=['id'], key='id', session=self.session, flush_session=True)
            widget = Widget()
            widget.id = 1
            widget, original = importer.update_object(widget, {'id': 1}), widget
            self.assertIs(widget, original)

        def test_flush_create_update(self):
            importer = saimport.ToSQLAlchemy(self.config, fields=['id'],
                                             session=self.session)
            with patch.object(self.session, 'flush') as flush:
                importer.flush_create_update()
                flush.assert_called_once_with()



# -*- coding: utf-8; -*-

import datetime
import shutil
import tempfile
from unittest import TestCase

from rattail.config import RattailConfig
from rattail.exceptions import ConfigurationError
from rattail.time import localtime

try:
    import sqlalchemy as sa
    from sqlalchemy import orm
    from rattail.db import Session, model
    from rattail.importing import postgresql as pgimport
    from .. import RattailTestCase, NullProgress
    from . import ImporterTester
    from .test_rattail import DualRattailTestCase
except ImportError:
    pass
else:

    class Widget(object):
        pass


    class TestBulkToPostgreSQL(TestCase):

        def setUp(self):
            self.tempdir = tempfile.mkdtemp()
            self.config = self.make_config()

        def tearDown(self):
            shutil.rmtree(self.tempdir)

        def make_config(self, workdir=True):
            cfg = RattailConfig()
            if workdir:
                cfg.setdefault('rattail', 'workdir', self.tempdir)
            cfg.setdefault('rattail', 'timezone.default', 'America/Chicago')
            return cfg

        def make_importer(self, **kwargs):
            kwargs.setdefault('config', self.config)
            kwargs.setdefault('fields', ['id']) # hack
            return pgimport.BulkToPostgreSQL(**kwargs)

        def test_data_path_property(self):
            self.config = self.make_config(workdir=False)
            self.config.setdefault('rattail', 'workdir', '/tmp')
            importer = pgimport.BulkToPostgreSQL(config=self.config, fields=['id'])

            # path leverages model name, so default is None
            self.assertEqual(importer.data_path, '/tmp/import_bulk_postgresql_None.csv')

            # but it will reflect model name
            importer.model_name = 'Foo'
            self.assertEqual(importer.data_path, '/tmp/import_bulk_postgresql_Foo.csv')

        def test_setup(self):
            importer = self.make_importer()
            self.assertFalse(hasattr(importer, 'data_buffer'))
            importer.setup()
            self.assertIsNotNone(importer.data_buffer)
            importer.data_buffer.close()

        def test_teardown(self):
            importer = self.make_importer()
            importer.data_buffer = open(importer.data_path, 'wb')
            importer.teardown()
            self.assertIsNone(importer.data_buffer)

        def test_prep_value_for_postgres(self):
            importer = self.make_importer()

            # constants
            self.assertEqual(importer.prep_value_for_postgres(None), '\\N')
            self.assertEqual(importer.prep_value_for_postgres(True), 't')
            self.assertEqual(importer.prep_value_for_postgres(False), 'f')

            # datetime (local zone is Chicago/CDT; UTC-5)
            value = localtime(self.config, datetime.datetime(2016, 5, 13, 12))
            self.assertEqual(importer.prep_value_for_postgres(value), '2016-05-13 17:00:00')

            # strings...

            # backslash is escaped by doubling
            self.assertEqual(importer.prep_value_for_postgres('\\'), '\\\\')

            # newlines are collapsed (\r\n -> \n) and escaped
            self.assertEqual(importer.prep_value_for_postgres('one\rtwo\nthree\r\nfour\r\nfive\nsix\rseven'), 'one\\rtwo\\nthree\\r\\nfour\\r\\nfive\\nsix\\rseven')

        def test_prep_data_for_postgres(self):
            importer = self.make_importer()
            time = localtime(self.config, datetime.datetime(2016, 5, 13, 12))
            data = {
                'none': None,
                'true': True,
                'false': False,
                'datetime': time,
                'backslash': '\\',
                'newlines': 'one\rtwo\nthree\r\nfour\r\nfive\nsix\rseven',
            }
            data = importer.prep_data_for_postgres(data)
            self.assertEqual(data['none'], '\\N')
            self.assertEqual(data['true'], 't')
            self.assertEqual(data['false'], 'f')
            self.assertEqual(data['datetime'], '2016-05-13 17:00:00')
            self.assertEqual(data['backslash'], '\\\\')
            self.assertEqual(data['newlines'], 'one\\rtwo\\nthree\\r\\nfour\\r\\nfive\\nsix\\rseven')


    ######################################################################
    # fake importer class, tested mostly for basic coverage
    ######################################################################

    class MockBulkImporter(pgimport.BulkToPostgreSQL):
        model_class = model.Department
        key = 'uuid'

        def normalize_local_object(self, obj):
            return obj

        def update_object(self, obj, host_data, local_data=None):
            return host_data


    class TestMockBulkImporter(DualRattailTestCase, ImporterTester):
        importer_class = MockBulkImporter

        sample_data = {
            1: {'number': 1, 'name': "Grocery", 'uuid': 'decd909a194011e688093ca9f40bc550'},
            2: {'number': 2, 'name': "Bulk", 'uuid': 'e633d54c194011e687e33ca9f40bc550'},
            3: {'number': 3, 'name': "HBA", 'uuid': 'e2bad79e194011e6a4783ca9f40bc550'},
        }

        def setUp(self):
            self.setup_rattail()
            self.tempdir = tempfile.mkdtemp()
            self.config.setdefault('rattail', 'workdir', self.tempdir)
            self.importer = self.make_importer()

        def tearDown(self):
            self.teardown_rattail()
            shutil.rmtree(self.tempdir)

        def make_importer(self, **kwargs):
            kwargs.setdefault('config', self.config)
            return super(TestMockBulkImporter, self).make_importer(**kwargs)

        def import_data(self, **kwargs):
            self.importer.session = self.session
            self.importer.host_session = self.host_session
            self.result = self.importer.import_data(**kwargs)

        def assert_import_created(self, *keys):
            pass

        def assert_import_updated(self, *keys):
            pass

        def assert_import_deleted(self, *keys):
            pass

        def test_create(self):
            if self.postgresql():
                with self.host_data(self.sample_data):
                    self.import_data()
                self.assert_import_created(3)

        def test_create_empty(self):
            if self.postgresql():
                with self.host_data({}):
                    self.import_data()
                self.assert_import_created(0)

        def test_max_create(self):
            if self.postgresql():
                with self.host_data(self.sample_data):
                    with self.local_data({}):
                        self.import_data(max_create=1)
                self.assert_import_created(1)

        def test_max_total_create(self):
            if self.postgresql():
                with self.host_data(self.sample_data):
                    with self.local_data({}):
                        self.import_data(max_total=1)
                self.assert_import_created(1)

        # # TODO: a bit hacky, leveraging the fact that 'user' is a reserved word
        # def test_table_name_is_reserved_word(self):
        #     if self.postgresql():
        #         from rattail.importing.rattail_bulk import UserImporter
        #         data = {
        #             '521a788e195911e688c13ca9f40bc550': {
        #                 'uuid': '521a788e195911e688c13ca9f40bc550',
        #                 'username': 'fred',
        #                 'active': True,
        #             },
        #         }
        #         self.importer = UserImporter(config=self.config)
        #         # with self.host_data(data):
        #         self.import_data(host_data=data)
        #         # self.assert_import_created(3)



# -*- coding: utf-8; -*-

from unittest import TestCase
from unittest.mock import patch, DEFAULT, Mock, MagicMock, call

from rattail.config import RattailConfig

try:
    from sqlalchemy import orm
    from rattail import db
    from rattail.db import changes, model
    from .. import DataTestCase
except ImportError:
    pass
else:

    class TestRecordChangesFunc(TestCase):

        def setUp(self):
            self.config = RattailConfig()
            self.app = self.config.get_app()

        def test_session_class(self):
            Session = orm.sessionmaker()
            if hasattr(Session, 'kw'):
                self.assertRaises(KeyError, Session.kw.__getitem__, 'rattail_record_changes')
            self.assertRaises(AttributeError, getattr, Session, 'rattail_record_changes')
            changes.record_changes(Session)
            self.assertTrue(Session.rattail_record_changes)

        def test_session_instance(self):
            session = db.Session()
            self.assertFalse(session.rattail_record_changes)
            changes.record_changes(session)
            self.assertTrue(session.rattail_record_changes)
            session.close()

        def test_recorder(self):

            # no recorder
            session = db.Session()
            self.assertRaises(AttributeError, getattr, session, 'rattail_change_recorder')
            session.close()

            # default recorder
            session = db.Session()
            changes.record_changes(session)
            self.assertIs(type(session.rattail_change_recorder), changes.ChangeRecorder)
            session.close()

            # specify recorder instance
            recorder = changes.ChangeRecorder(self.config)
            session = db.Session()
            changes.record_changes(session, recorder=recorder)
            self.assertIs(session.rattail_change_recorder, recorder)
            session.close()

            # specify recorder factory
            session = db.Session()
            changes.record_changes(session, recorder=changes.ChangeRecorder, config=self.config)
            self.assertIs(type(session.rattail_change_recorder), changes.ChangeRecorder)
            session.close()

            # specify recorder spec via config
            config = RattailConfig()
            config.setdefault('rattail.db', 'changes.recorder', 'rattail.db.changes:ChangeRecorder')
            session = db.Session()
            changes.record_changes(session, config=config)
            self.assertIs(type(session.rattail_change_recorder), changes.ChangeRecorder)
            session.close()

            # invalid recorder
            session = db.Session()
            self.assertRaises(ValueError, changes.record_changes, session, recorder='bogus')
            session.close()


    class TestChangeRecorder(DataTestCase):

        def extra_setup(self):
            self.config = RattailConfig()
            self.app = self.config.get_app()

        def test_ignore_object(self):
            recorder = changes.ChangeRecorder(self.config)
            self.assertTrue(recorder.ignore_object(model.Setting()))
            self.assertTrue(recorder.ignore_object(model.Change()))
            self.assertTrue(recorder.ignore_object(model.DataSyncChange()))
            self.assertFalse(recorder.ignore_object(model.Product()))
            self.assertFalse(recorder.ignore_object(model.Customer()))

        def test_process_new_object(self):
            self.assertEqual(self.session.query(model.Change).count(), 0)
            recorder = changes.ChangeRecorder(self.config)
            product = model.Product(uuid='6de299ca178d11e6be2c3ca9f40bc550')
            recorder.process_new_object(self.session, product)
            change = self.session.query(model.Change).one()
            self.assertEqual(change.class_name, 'Product')
            self.assertEqual(change.object_key, '6de299ca178d11e6be2c3ca9f40bc550')
            self.assertFalse(change.deleted)

        def test_process_dirty_object(self):
            self.assertEqual(self.session.query(model.Change).count(), 0)
            recorder = changes.ChangeRecorder(self.config)
            product = model.Product(uuid='6de299ca178d11e6be2c3ca9f40bc550')
            recorder.process_dirty_object(self.session, product)
            change = self.session.query(model.Change).one()
            self.assertEqual(change.class_name, 'Product')
            self.assertEqual(change.object_key, '6de299ca178d11e6be2c3ca9f40bc550')
            self.assertFalse(change.deleted)

        def test_process_deleted_object(self):
            self.assertEqual(self.session.query(model.Change).count(), 0)
            recorder = changes.ChangeRecorder(self.config)
            product = model.Product(uuid='6de299ca178d11e6be2c3ca9f40bc550')
            recorder.process_deleted_object(self.session, product)
            change = self.session.query(model.Change).one()
            self.assertEqual(change.class_name, 'Product')
            self.assertEqual(change.object_key, '6de299ca178d11e6be2c3ca9f40bc550')
            self.assertTrue(change.deleted)

        def test_process_deleted_object_special(self):
            self.assertEqual(self.session.query(model.Change).count(), 0)
            recorder = changes.ChangeRecorder(self.config)

            person = model.Person(uuid='06100a34178e11e6a8633ca9f40bc550')
            for Model in (model.PersonEmailAddress, model.PersonPhoneNumber, model.PersonMailingAddress):

                self.assertEqual(self.session.query(model.Change).count(), 0)
                uuid = self.app.make_uuid()
                obj = Model(uuid=uuid, person=person)
                recorder.process_deleted_object(self.session, obj)
                self.assertEqual(self.session.query(model.Change).count(), 2)

                change = self.session.query(model.Change).filter_by(class_name=Model.__name__).one()
                self.assertEqual(change.object_key, uuid)
                self.assertTrue(change.deleted)

                change = self.session.query(model.Change).filter_by(class_name='Person').one()
                self.assertEqual(change.object_key, '06100a34178e11e6a8633ca9f40bc550')
                self.assertFalse(change.deleted)

                self.session.query(model.Change).delete(synchronize_session=False)

            employee = model.Employee(uuid='b302ac38178e11e6b8843ca9f40bc550')
            for Model in (model.EmployeeStore, model.EmployeeDepartment):

                self.assertEqual(self.session.query(model.Change).count(), 0)
                uuid = self.app.make_uuid()
                obj = Model(uuid=uuid, employee=employee)
                recorder.process_deleted_object(self.session, obj)
                self.assertEqual(self.session.query(model.Change).count(), 2)

                change = self.session.query(model.Change).filter_by(class_name=Model.__name__).one()
                self.assertEqual(change.object_key, uuid)
                self.assertTrue(change.deleted)

                change = self.session.query(model.Change).filter_by(class_name='Employee').one()
                self.assertEqual(change.object_key, 'b302ac38178e11e6b8843ca9f40bc550')
                self.assertFalse(change.deleted)

                self.session.query(model.Change).delete(synchronize_session=False)

        def test_record_change(self):
            self.assertEqual(self.session.query(model.Change).count(), 0)
            recorder = changes.ChangeRecorder(self.config)

            recorder.record_change(self.session, class_name='Bogus', object_key='bogus', deleted=False)
            self.assertEqual(self.session.query(model.Change).count(), 1)
            change = self.session.query(model.Change).one()
            self.assertEqual(change.class_name, 'Bogus')
            self.assertEqual(change.object_key, 'bogus')
            self.assertFalse(change.deleted)

            recorder.record_change(self.session, class_name='Invalid', object_key='invalid', deleted=True)
            self.assertEqual(self.session.query(model.Change).count(), 2)
            change = self.session.query(model.Change).filter_by(class_name='Invalid').one()
            self.assertEqual(change.object_key, 'invalid')
            self.assertTrue(change.deleted)

        def test_record_rattail_change(self):
            self.assertEqual(self.session.query(model.Change).count(), 0)
            recorder = changes.ChangeRecorder(self.config)

            # ignore change object (TODO: redundant?)
            self.assertFalse(recorder.record_rattail_change(self.session, model.Change()))
            self.assertFalse(recorder.record_rattail_change(self.session, model.DataSyncChange()))
            self.assertEqual(self.session.query(model.Change).count(), 0)

            # ignore batch object (TODO: redundant?)
            self.assertFalse(recorder.record_rattail_change(self.session, model.BatchMixin()))
            self.assertFalse(recorder.record_rattail_change(self.session, model.BatchRowMixin()))
            self.assertEqual(self.session.query(model.Change).count(), 0)

            # ignore instance with no UUID attribute
            self.assertFalse(recorder.record_rattail_change(self.session, model.Setting()))
            self.assertFalse(recorder.record_rattail_change(self.session, model.Permission()))
            self.assertEqual(self.session.query(model.Change).count(), 0)

            product = model.Product(uuid='c1b0fb94178f11e680fd3ca9f40bc550')

            # default
            self.assertTrue(recorder.record_rattail_change(self.session, product))
            change = self.session.query(model.Change).one()
            self.assertEqual(change.class_name, 'Product')
            self.assertEqual(change.object_key, 'c1b0fb94178f11e680fd3ca9f40bc550')
            self.assertFalse(change.deleted)
            self.session.query(model.Change).delete(synchronize_session=False)

            # new
            self.assertTrue(recorder.record_rattail_change(self.session, product, type_='new'))
            change = self.session.query(model.Change).one()
            self.assertEqual(change.class_name, 'Product')
            self.assertEqual(change.object_key, 'c1b0fb94178f11e680fd3ca9f40bc550')
            self.assertFalse(change.deleted)
            self.session.query(model.Change).delete(synchronize_session=False)

            # dirty
            self.assertTrue(recorder.record_rattail_change(self.session, product, type_='dirty'))
            change = self.session.query(model.Change).one()
            self.assertEqual(change.class_name, 'Product')
            self.assertEqual(change.object_key, 'c1b0fb94178f11e680fd3ca9f40bc550')
            self.assertFalse(change.deleted)
            self.session.query(model.Change).delete(synchronize_session=False)

            # deleted
            self.assertTrue(recorder.record_rattail_change(self.session, product, type_='deleted'))
            change = self.session.query(model.Change).one()
            self.assertEqual(change.class_name, 'Product')
            self.assertEqual(change.object_key, 'c1b0fb94178f11e680fd3ca9f40bc550')
            self.assertTrue(change.deleted)
            self.session.query(model.Change).delete(synchronize_session=False)


    class TestChangeRecorderLegacy(TestCase):

        def setUp(self):
            self.config = RattailConfig()

        def test_init(self):
            recorder = changes.ChangeRecorder(self.config)

        # def test_record_change(self):
        #     session = Mock()
        #     recorder = changes.ChangeRecorder()
        #     recorder.ensure_uuid = Mock()

        #     # don't record changes for changes
        #     self.assertFalse(recorder.record_change(session, model.Change()))

        #     # don't record changes for objects with no uuid attribute
        #     self.assertFalse(recorder.record_change(session, object()))

        #     # none of the above should have involved a call to `ensure_uuid()`
        #     self.assertFalse(recorder.ensure_uuid.called)

        #     # so far no *new* changes have been created
        #     self.assertFalse(session.add.called)

        #     # mock up session to force new change creation
        #     session.query.return_value = session
        #     session.get.return_value = None
        #     self.assertTrue(recorder.record_change(session, model.Product()))

        @patch.multiple('rattail.db.changes', object_mapper=DEFAULT)
        def test_ensure_uuid(self, object_mapper):
            recorder = changes.ChangeRecorder(self.config)
            uuid_column = Mock()
            object_mapper.return_value.columns.__getitem__.return_value = uuid_column

            with patch.object(recorder.app, 'make_uuid') as make_uuid:

                # uuid already present
                product = model.Product(uuid='some_uuid')
                recorder.ensure_uuid(product)
                self.assertEqual(product.uuid, 'some_uuid')
                self.assertFalse(make_uuid.called)

                # no uuid yet, auto-generate
                uuid_column.foreign_keys = False
                make_uuid.return_value = 'another_uuid'
                product = model.Product()
                self.assertTrue(product.uuid is None)
                recorder.ensure_uuid(product)
                make_uuid.assert_called_once_with()
                self.assertEqual(product.uuid, 'another_uuid')

                # some heavy mocking for following tests
                uuid_column.foreign_keys = True
                remote_side = MagicMock(key='uuid')
                prop = MagicMock(__class__=orm.RelationshipProperty, key='foreign_thing')
                prop.remote_side.__len__.return_value = 1
                prop.remote_side.__iter__.return_value = [remote_side]
                object_mapper.return_value.iterate_properties.__iter__.return_value = [prop]

                # uuid fetched from existing foreign key object
                make_uuid.reset_mock()
                instance = Mock(uuid=None, foreign_thing=Mock(uuid='secondary_uuid'))
                recorder.ensure_uuid(instance)
                self.assertFalse(make_uuid.called)
                self.assertEqual(instance.uuid, 'secondary_uuid')

                # foreign key object doesn't exist; uuid generated as fallback
                make_uuid.return_value = 'fallback_uuid'
                instance = Mock(uuid=None, foreign_thing=None)
                recorder.ensure_uuid(instance)
                make_uuid.assert_called_once_with()
                self.assertEqual(instance.uuid, 'fallback_uuid')


    class TestFunctionalChanges(DataTestCase):

        def setUp(self):
            super().setUp()
            self.config = RattailConfig()
            changes.record_changes(self.session, config=self.config)

        def test_add(self):
            product = model.Product()
            self.session.add(product)
            self.session.commit()

            self.assertEqual(self.session.query(model.Change).count(), 1)
            change = self.session.query(model.Change).one()
            self.assertEqual(change.class_name, 'Product')
            self.assertEqual(change.instance_uuid, product.uuid)
            self.assertFalse(change.deleted)

        def test_change(self):
            product = model.Product()
            self.session.add(product)
            self.session.commit()

            self.assertEqual(self.session.query(model.Change).count(), 1)
            self.session.query(model.Change).delete()
            self.assertEqual(self.session.query(model.Change).count(), 0)

            product.description = 'Acme Bricks'
            self.session.commit()

            self.assertEqual(self.session.query(model.Change).count(), 1)
            change = self.session.query(model.Change).one()
            self.assertEqual(change.class_name, 'Product')
            self.assertEqual(change.instance_uuid, product.uuid)
            self.assertFalse(change.deleted)

        def test_delete(self):
            product = model.Product()
            self.session.add(product)
            self.session.commit()

            self.assertEqual(self.session.query(model.Change).count(), 1)
            self.session.query(model.Change).delete()
            self.assertEqual(self.session.query(model.Change).count(), 0)

            self.session.delete(product)

            self.assertEqual(self.session.query(model.Change).count(), 1)
            change = self.session.query(model.Change).one()
            self.assertEqual(change.class_name, 'Product')
            self.assertEqual(change.instance_uuid, product.uuid)
            self.assertTrue(change.deleted)

        def test_orphan_change(self):
            department = model.Department()
            subdepartment = model.Subdepartment()
            department.subdepartments.append(subdepartment)
            self.session.add(department)
            self.session.commit()

            self.assertEqual(self.session.query(model.Change).count(), 2)
            change = self.session.query(model.Change).filter_by(class_name='Department').one()
            self.assertFalse(change.deleted)
            change = self.session.query(model.Change).filter_by(class_name='Subdepartment').one()
            self.assertFalse(change.deleted)

            self.session.query(model.Change).delete()
            self.assertEqual(self.session.query(model.Change).count(), 0)

            # Creating an orphaned Subdepartment, which should be recorded as a
            # *change* due to the cascade rules in effect.
            department.subdepartments.remove(subdepartment)
            self.session.commit()

            self.assertEqual(self.session.query(model.Change).count(), 2)
            change = self.session.query(model.Change).filter_by(class_name='Department').one()
            self.assertFalse(change.deleted)
            change = self.session.query(model.Change).filter_by(class_name='Subdepartment').one()
            self.assertFalse(change.deleted)
            self.assertEqual(self.session.query(model.Subdepartment).count(), 1)

        def test_orphan_delete(self):
            customer = model.Customer()
            group = model.CustomerGroup()
            customer.groups.append(group)
            self.session.add(customer)
            self.session.commit()

            self.assertEqual(self.session.query(model.Change).count(), 3)
            change = self.session.query(model.Change).filter_by(class_name='Customer').one()
            self.assertFalse(change.deleted)
            change = self.session.query(model.Change).filter_by(class_name='CustomerGroup').one()
            self.assertFalse(change.deleted)
            change = self.session.query(model.Change).filter_by(class_name='CustomerGroupAssignment').one()
            self.assertFalse(change.deleted)

            self.session.query(model.Change).delete()
            self.assertEqual(self.session.query(model.Change).count(), 0)

            # Creating an orphaned CustomerGroupAssociation, which should be
            # recorded as a *deletion* due to the cascade rules in effect.  Note
            # that the CustomerGroup is not technically an orphan and in fact is
            # not even changed.
            customer.groups.remove(group)
            self.session.commit()

            self.assertEqual(self.session.query(model.Change).count(), 2)
            change = self.session.query(model.Change).filter_by(class_name='Customer').one()
            self.assertFalse(change.deleted)
            change = self.session.query(model.Change).filter_by(class_name='CustomerGroupAssignment').one()
            self.assertTrue(change.deleted)
            self.assertEqual(self.session.query(model.CustomerGroupAssignment).count(), 0)



# -*- coding: utf-8; -*-

import os
from unittest import TestCase
from unittest.mock import Mock

from rattail.config import RattailConfig

try:
    import sqlalchemy as sa
    from rattail.db import model, Session, ConfigExtension
    from rattail.importing import model as import_model
    from .. import RattailTestCase
except ImportError:
    pass
else:

    class TestAdminUser(TestCase):

        def setUp(self):
            self.config = RattailConfig(defaults={
                'rattail.timezone.default': 'America/Chicago',
            })
            engine_url = os.environ.get('RATTAIL_TEST_ENGINE_URL', 'sqlite://')
            self.engine = sa.create_engine(engine_url)
            model.Base.metadata.create_all(bind=self.engine)
            Session.configure(bind=self.engine)
            self.app = self.config.get_app()
            self.session = self.app.make_session()

        def tearDown(self):
            self.session.close()
            model.Base.metadata.drop_all(bind=self.engine)
            Session.configure(bind=None)

        def make_importer(self, **kwargs):
            kwargs.setdefault('config', self.config)
            kwargs.setdefault('session', self.session)
            return import_model.AdminUserImporter(**kwargs)

        def get_admin(self):
            auth = self.app.get_auth_handler()
            return auth.get_role_administrator(self.session)

        def test_supported_fields(self):
            importer = import_model.UserImporter(self.config)
            standard_fields = importer.fields
            importer = self.make_importer()
            extra_fields = set(importer.fields) - set(standard_fields)
            self.assertEqual(len(extra_fields), 1)
            self.assertEqual(list(extra_fields)[0], 'admin')

        def test_normalize_local_object(self):
            importer = self.make_importer()
            importer.setup()

            user = model.User()
            user.username = 'fred'
            self.session.add(user)
            self.session.flush()

            data = importer.normalize_local_object(user)
            self.assertFalse(data['admin'])

            user.roles.append(self.get_admin())
            self.session.flush()
            data = importer.normalize_local_object(user)
            self.assertTrue(data['admin'])

        def test_update_object(self):
            importer = self.make_importer(fields=['uuid', 'admin'])
            data = {'uuid': 'ccb1915419e511e6a3ad3ca9f40bc550'}
            user = model.User(**data)
            admin = self.get_admin()
            self.assertNotIn(admin, user.roles)

            data['admin'] = True
            importer.update_object(user, data)
            self.assertIn(admin, user.roles)

            data['admin'] = False
            importer.update_object(user, data)
            self.assertNotIn(admin, user.roles)



# -*- coding: utf-8; -*-

import os
import shutil
import decimal
from unittest import TestCase

from rattail.config import make_config
from rattail.excel import ExcelWriter
from rattail.gpc import GPC

try:
    import sqlalchemy as sa
    from rattail.batch import vendorcatalog as mod
    from rattail.db import Session
except ImportError:
    pass
else:

    class TestProductBatchHandler(TestCase):

        def setUp(self):
            self.config = self.make_config()
            self.app = self.config.get_app()
            self.handler = self.make_handler()

        def make_config(self):
            return make_config([], extend=False)

        def make_handler(self):
            return mod.VendorCatalogHandler(self.config)

        def test_allow_future(self):

            # off by default
            result = self.handler.allow_future()
            self.assertFalse(result)

            # but can be enabled via config
            self.config.setdefault('rattail.batch', 'vendor_catalog.allow_future',
                                   'true')
            result = self.handler.allow_future()
            self.assertTrue(result)

        def test_populate_from_file(self):
            engine = sa.create_engine('sqlite://')
            model = self.app.model
            model.Base.metadata.create_all(bind=engine)
            session = Session(bind=engine)
            app = self.config.get_app()

            # we'll need a user to create the batches
            user = model.User(username='ralph')
            session.add(user)

            # make root folder to contain all temp files
            tempdir = app.make_temp_dir()

            # generate sample xlsx file
            path = os.path.join(tempdir, 'sample.xlsx')
            writer = ExcelWriter(path, ['UPC', 'Vendor Code', 'Unit Cost'])
            writer.write_header()
            writer.write_row(['074305001321', '123456', 4.19], row=2)
            writer.save()

            # make, configure folder for batch files
            filesdir = os.path.join(tempdir, 'batch_files')
            os.makedirs(filesdir)
            self.config.setdefault('rattail', 'batch.files', filesdir)

            # make the basic batch
            batch = model.VendorCatalogBatch(uuid=app.make_uuid(),
                                             id=1, created_by=user)
            session.add(batch)

            # batch must have certain attributes, else error
            self.assertRaises(ValueError, self.handler.populate_from_file, batch)
            self.handler.set_input_file(batch, path) # sets batch.filename
            self.assertRaises(ValueError, self.handler.populate_from_file, batch)
            batch.parser_key = 'rattail.contrib.generic'

            # and finally, test our method proper
            self.handler.setup_populate(batch)
            self.handler.populate_from_file(batch)
            self.assertEqual(len(batch.data_rows), 1)
            row = batch.data_rows[0]
            self.assertEqual(row.item_entry, '074305001321')
            self.assertEqual(row.vendor_code, '123456')
            self.assertEqual(row.unit_cost, decimal.Decimal('4.19'))

            shutil.rmtree(tempdir)
            session.rollback()
            session.close()

        def test_identify_product(self):
            engine = sa.create_engine('sqlite://')
            model = self.app.model
            model.Base.metadata.create_all(bind=engine)
            session = Session(bind=engine)
            app = self.config.get_app()

            # make a test user, vendor, product, cost
            user = model.User(username='ralph')
            session.add(user)
            vendor = model.Vendor()
            session.add(vendor)
            product = model.Product(upc=GPC('074305001321'))
            session.add(product)
            cost = model.ProductCost(vendor=vendor,
                                     code='123456',
                                     case_size=12,
                                     case_cost=decimal.Decimal('54.00'),
                                     unit_cost=decimal.Decimal('4.50'))
            product.costs.append(cost)

            # also a batch to contain the rows
            batch = model.VendorCatalogBatch(uuid=app.make_uuid(),
                                             id=1, created_by=user,
                                             vendor=vendor,
                                             filename='sample.xlsx',
                                             parser_key='rattail.contrib.generic')
            session.add(batch)

            # row w/ no interesting attributes cannot yield a product
            row = model.VendorCatalogBatchRow()
            batch.data_rows.append(row)
            result = self.handler.identify_product(row)
            self.assertIsNone(result)

            # but if we give row a upc, product is found
            row.upc = GPC('074305001321')
            result = self.handler.identify_product(row)
            self.assertIs(result, product)

            # now try one with vendor code instead of upc
            row = model.VendorCatalogBatchRow(vendor_code='123456')
            batch.data_rows.append(row)
            result = self.handler.identify_product(row)
            self.assertIs(result, product)

            session.rollback()
            session.close()

        def test_refresh_row(self):
            engine = sa.create_engine('sqlite://')
            model = self.app.model
            model.Base.metadata.create_all(bind=engine)
            session = Session(bind=engine)
            app = self.config.get_app()

            # make a test user, vendor, product
            user = model.User(username='ralph')
            session.add(user)
            vendor = model.Vendor()
            session.add(vendor)
            product = model.Product(upc=GPC('074305001321'))
            session.add(product)

            # also a batch to contain the rows
            batch = model.VendorCatalogBatch(uuid=app.make_uuid(),
                                             id=1, created_by=user,
                                             vendor=vendor,
                                             filename='sample.xlsx',
                                             parser_key='rattail.contrib.generic')
            session.add(batch)

            # empty row is just marked as product not found
            row = model.VendorCatalogBatchRow()
            batch.data_rows.append(row)
            self.handler.refresh_row(row)
            self.assertEqual(row.status_code, row.STATUS_PRODUCT_NOT_FOUND)

            # row with upc is matched with product; also make sure unit
            # cost is calculated from case cost
            row = model.VendorCatalogBatchRow(upc=GPC('074305001321'),
                                              case_size=12,
                                              case_cost=decimal.Decimal('58.00'))
            batch.data_rows.append(row)
            self.handler.refresh_row(row)
            self.assertIs(row.product, product)
            self.assertEqual(row.status_code, row.STATUS_NEW_COST)
            self.assertEqual(row.case_cost, 58)
            self.assertEqual(row.case_size, 12)
            self.assertEqual(row.unit_cost, decimal.Decimal('4.8333'))

            # now we add a cost to the master product, and make sure new
            # row will reflect an update for that cost
            cost = model.ProductCost(vendor=vendor, 
                                     case_size=12,
                                     case_cost=decimal.Decimal('54.00'),
                                     unit_cost=decimal.Decimal('4.50'))
            product.costs.append(cost)
            row = model.VendorCatalogBatchRow(upc=GPC('074305001321'),
                                              case_size=12,
                                              case_cost=decimal.Decimal('58.00'))
            batch.data_rows.append(row)
            self.handler.refresh_row(row)
            self.assertIs(row.product, product)
            self.assertEqual(row.status_code, row.STATUS_CHANGE_COST)
            self.assertEqual(row.old_case_cost, 54)
            self.assertEqual(row.case_cost, 58)
            self.assertEqual(row.old_unit_cost, decimal.Decimal('4.50'))
            self.assertEqual(row.unit_cost, decimal.Decimal('4.8333'))

            # and finally let's refresh everything, note that row #2
            # should now *also* get "change cost" status
            row = batch.data_rows[1]
            self.assertEqual(row.status_code, row.STATUS_NEW_COST)
            self.handler.setup_refresh(batch)
            for row in batch.data_rows:
                self.handler.refresh_row(row)
            self.assertEqual(row.status_code, row.STATUS_CHANGE_COST)

            session.rollback()
            session.close()



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2024 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
DEPRECATED: Authentication & Authorization

This entire module has been deprecated; please use the :term:`auth
handler` instead.
"""

import warnings

from passlib.context import CryptContext

from rattail.db import model


password_context = CryptContext(schemes=['bcrypt'])


def authenticate_user(session, userobj, password):
    """
    DEPRECATED; use
    :meth:`wuttjamaican:wuttjamaican.auth.AuthHandler.authenticate_user()`
    instead.
    """
    warnings.warn("authenticate_user() function is deprecated; "
                  "please use AuthHandler.authenticate_user() instead",
                  DeprecationWarning, stacklevel=2)

    if isinstance(userobj, model.User):
        user = userobj
    else:
        user = session.query(model.User)\
                      .filter_by(username=userobj)\
                      .first()

    if user and user.active and user.password is not None:
        if password_context.verify(password, user.password):
            return user


def set_user_password(user, password):
    """
    DEPRECATED; use
    :meth:`wuttjamaican:wuttjamaican.auth.AuthHandler.set_user_password()`
    instead.
    """
    warnings.warn("set_user_password() function is deprecated; "
                  "please use AuthHandler.set_user_password() instead",
                  DeprecationWarning, stacklevel=2)

    user.password = password_context.hash(password)


def special_role(session, uuid, name):
    """ """
    warnings.warn("special_role() function is deprecated; "
                  "please use AuthHandler._special_role() instead",
                  DeprecationWarning, stacklevel=2)

    role = session.get(model.Role, uuid)
    if not role:
        role = model.Role(uuid=uuid, name=name)
        session.add(role)
    return role


def administrator_role(session):
    """
    DEPRECATED; use
    :meth:`wuttjamaican:wuttjamaican.auth.AuthHandler.get_role_administrator()`
    instead.
    """
    warnings.warn("administrator_role() function is deprecated; "
                  "please use AuthHandler.get_role_administrator() instead",
                  DeprecationWarning, stacklevel=2)

    return special_role(session, 'd937fa8a965611dfa0dd001143047286', 'Administrator')


def guest_role(session):
    """
    DEPRECATED; use
    :meth:`wuttjamaican:wuttjamaican.auth.AuthHandler.get_role_anonymous()`
    instead.
    """
    warnings.warn("guest_role() function is deprecated; "
                  "please use AuthHandler.get_role_anonymous() instead",
                  DeprecationWarning, stacklevel=2)

    return special_role(session, 'f8a27c98965a11dfaff7001143047286', 'Guest')


def authenticated_role(session):
    """
    DEPRECATED; use
    :meth:`wuttjamaican:wuttjamaican.auth.AuthHandler.get_role_authenticated()`
    instead.
    """
    warnings.warn("authenticated_role() function is deprecated; "
                  "please use AuthHandler.get_role_authenticated() instead",
                  DeprecationWarning, stacklevel=2)

    return special_role(session, 'b765a9cc331a11e6ac2a3ca9f40bc550', "Authenticated")


def grant_permission(role, permission):
    """
    DEPRECATED; use
    :meth:`wuttjamaican:wuttjamaican.auth.AuthHandler.grant_permission()`
    instead.
    """
    warnings.warn("grant_permission() function is deprecated; "
                  "please use AuthHandler.grant_permission() instead",
                  DeprecationWarning, stacklevel=2)

    # TODO: Make this a `Role` method (or make `Role.permissions` a `set` so we
    # can do `role.permissions.add('some.perm')` ?).
    if permission not in role.permissions:
        role.permissions.append(permission)


def revoke_permission(role, permission):
    """
    DEPRECATED; use
    :meth:`wuttjamaican:wuttjamaican.auth.AuthHandler.revoke_permission()`
    instead.
    """
    warnings.warn("revoke_permission() function is deprecated; "
                  "please use AuthHandler.revoke_permission() instead",
                  DeprecationWarning, stacklevel=2)

    if permission in role.permissions:
        role.permissions.remove(permission)


def has_permission(session, principal, permission, include_guest=True, include_authenticated=True):
    """
    DEPRECATED; use
    :meth:`wuttjamaican:wuttjamaican.auth.AuthHandler.has_permission()`
    instead.
    """
    warnings.warn("has_permission() function is deprecated; "
                  "please use AuthHandler.has_permission() instead",
                  DeprecationWarning, stacklevel=2)

    if hasattr(principal, 'roles'):
        roles = list(principal.roles)
        if include_authenticated:
            roles.append(authenticated_role(session))
    elif principal is not None:
        roles = [principal]
    else:
        roles = []

    if include_guest:
        roles.append(guest_role(session))
    for role in roles:
        for perm in role.permissions:
            if perm == permission:
                return True
    return False


def cache_permissions(session, principal, include_guest=True, include_authenticated=True):
    """
    DEPRECATED; use
    :meth:`wuttjamaican:wuttjamaican.auth.AuthHandler.get_permissions()`
    instead.
    """
    warnings.warn("cache_permissions() function is deprecated; "
                  "please use AuthHandler.get_permissions() instead",
                  DeprecationWarning, stacklevel=2)

    # we will use any `roles` attribute which may be present.  in practice we
    # would be assuming a User in this case
    if hasattr(principal, 'roles'):
        roles = list(principal.roles)

        # here our User assumption gets a little more explicit
        if include_authenticated:
            roles.append(authenticated_role(session))

    # otherwise a non-null principal is assumed to be a Role
    elif principal is not None:
        roles = [principal]

    # fallback assumption is "no roles"
    else:
        roles = []

    # maybe include guest roles
    if include_guest:
        roles.append(guest_role(session))

    # build the permissions cache
    cache = set()
    for role in roles:
        cache.update(role.permissions)

    return cache



# -*- coding: utf-8; -*-

import datetime
from unittest import TestCase
from unittest.mock import patch, Mock
from collections import OrderedDict

import pytz

from rattail.importing import handlers, Importer
from rattail.config import make_config
from . import ImporterTester
from .test_importers import MockImporter


class ImportHandlerBattery(ImporterTester):

    def make_config(self):
        config = make_config([], extend=False)
        config.setdefault('rattail', 'timezone.default', 'America/Chicago')
        return config

    def test_init(self):

        # vanilla
        handler = self.handler_class()
        self.assertEqual(handler.importers, {})
        self.assertEqual(handler.get_importers(), {})
        self.assertEqual(handler.get_importer_keys(), [])
        self.assertEqual(handler.get_default_keys(), [])
        self.assertFalse(handler.commit_host_partial)

        # with config
        handler = self.handler_class()
        self.assertIsNone(handler.config)
        config = self.make_config()
        handler = self.handler_class(config=config)
        self.assertIs(handler.config, config)

        # dry run
        handler = self.handler_class()
        self.assertFalse(handler.dry_run)
        handler = self.handler_class(dry_run=True)
        self.assertTrue(handler.dry_run)

        # extra kwarg
        handler = self.handler_class()
        self.assertRaises(AttributeError, getattr, handler, 'foo')
        handler = self.handler_class(foo='bar')
        self.assertEqual(handler.foo, 'bar')

    def test_get_importer(self):
        get_importers = Mock(return_value={'foo': Importer})

        # no importers
        handler = self.make_handler()
        self.assertIsNone(handler.get_importer('foo'))

        # with config
        config = self.make_config()
        with patch.object(self.handler_class, 'get_importers', get_importers):
            handler = self.handler_class(config)
        importer = handler.get_importer('foo')
        self.assertIs(type(importer), Importer)
        self.assertIs(importer.config, config)
        self.assertIs(importer.handler, handler)

        # # batch size
        # with patch.object(self.handler_class, 'get_importers', get_importers):
        #     handler = self.handler_class()
        # importer = handler.get_importer('foo')
        # self.assertEqual(importer.batch_size, 200)
        # with patch.object(self.handler_class, 'get_importers', get_importers):
        #     handler = self.handler_class(batch_size=42)
        # importer = handler.get_importer('foo')
        # self.assertEqual(importer.batch_size, 42)

        # dry run
        with patch.object(self.handler_class, 'get_importers', get_importers):
            handler = self.handler_class(config)
        importer = handler.get_importer('foo')
        self.assertFalse(importer.dry_run)
        with patch.object(self.handler_class, 'get_importers', get_importers):
            handler = self.handler_class(config, dry_run=True)
        importer = handler.get_importer('foo')
        self.assertTrue(handler.dry_run)

        # host title
        with patch.object(self.handler_class, 'get_importers', get_importers):
            handler = self.handler_class(config)
        importer = handler.get_importer('foo')
        self.assertIsNone(importer.host_system_title)
        handler.host_title = "Foo"
        importer = handler.get_importer('foo')
        self.assertEqual(importer.host_system_title, "Foo")

        # extra kwarg
        with patch.object(self.handler_class, 'get_importers', get_importers):
            handler = self.handler_class(config)
        importer = handler.get_importer('foo')
        self.assertRaises(AttributeError, getattr, importer, 'bar')
        importer = handler.get_importer('foo', bar='baz')
        self.assertEqual(importer.bar, 'baz')

    def test_get_importer_kwargs(self):

        # empty by default
        handler = self.make_handler()
        self.assertEqual(handler.get_importer_kwargs('foo'), {})

        # extra kwargs are preserved
        handler = self.make_handler()
        self.assertEqual(handler.get_importer_kwargs('foo', bar='baz'), {'bar': 'baz'})

    def test_begin_transaction(self):
        handler = self.make_handler()
        with patch.object(handler, 'begin_host_transaction') as begin_host:
            with patch.object(handler, 'begin_local_transaction') as begin_local:
                handler.begin_transaction()
                begin_host.assert_called_once_with()
                begin_local.assert_called_once_with()

    def test_begin_host_transaction(self):
        handler = self.make_handler()
        handler.begin_host_transaction()

    def test_begin_local_transaction(self):
        handler = self.make_handler()
        handler.begin_local_transaction()

    def test_commit_transaction(self):
        handler = self.make_handler()
        with patch.object(handler, 'commit_host_transaction') as commit_host:
            with patch.object(handler, 'commit_local_transaction') as commit_local:
                handler.commit_transaction()
                commit_host.assert_called_once_with()
                commit_local.assert_called_once_with()

    def test_commit_host_transaction(self):
        handler = self.make_handler()
        handler.commit_host_transaction()

    def test_commit_local_transaction(self):
        handler = self.make_handler()
        handler.commit_local_transaction()

    def test_rollback_transaction(self):
        handler = self.make_handler()
        with patch.object(handler, 'rollback_host_transaction') as rollback_host:
            with patch.object(handler, 'rollback_local_transaction') as rollback_local:
                handler.rollback_transaction()
                rollback_host.assert_called_once_with()
                rollback_local.assert_called_once_with()

    def test_rollback_host_transaction(self):
        handler = self.make_handler()
        handler.rollback_host_transaction()

    def test_rollback_local_transaction(self):
        handler = self.make_handler()
        handler.rollback_local_transaction()

    def test_import_data(self):
        self.config = self.make_config()
        handler = self.make_handler()
        result = handler.import_data()
        self.assertEqual(result, {})

    def test_import_data_dry_run(self):
        self.config = self.make_config()

        # as init kwarg
        handler = self.make_handler(dry_run=True)
        with patch.object(handler, 'commit_transaction') as commit:
            with patch.object(handler, 'rollback_transaction') as rollback:
                handler.import_data()
                self.assertFalse(commit.called)
                rollback.assert_called_once_with()
        self.assertTrue(handler.dry_run)

        # as import kwarg
        handler = self.make_handler()
        with patch.object(handler, 'commit_transaction') as commit:
            with patch.object(handler, 'rollback_transaction') as rollback:
                handler.import_data(dry_run=True)
                self.assertFalse(commit.called)
                rollback.assert_called_once_with()
        self.assertTrue(handler.dry_run)

    def test_import_data_invalid_model(self):
        handler = self.make_handler()
        importer = Mock()
        importer.import_data.return_value = [], [], []
        FooImporter = Mock(return_value=importer)
        handler.importers = {'Foo': FooImporter}

        handler.import_data('Foo')
        self.assertEqual(FooImporter.call_count, 1)
        importer.import_data.assert_called_once_with()

        FooImporter.reset_mock()
        importer.reset_mock()

        handler.import_data('Missing')
        self.assertFalse(FooImporter.called)
        self.assertFalse(importer.called)

    def test_import_data_with_changes(self):
        handler = self.make_handler()
        importer = Mock()
        FooImporter = Mock(return_value=importer)
        handler.importers = {'Foo': FooImporter}

        importer.import_data.return_value = [], [], []
        with patch.object(handler, 'process_changes') as process:
            handler.import_data('Foo')
            self.assertFalse(process.called)

        importer.import_data.return_value = [1], [2], [3]
        with patch.object(handler, 'process_changes') as process:
            handler.import_data('Foo')
            process.assert_called_once_with({'Foo': ([1], [2], [3])})

    def test_import_data_commit_host_partial(self):
        self.config = self.make_config()
        handler = self.make_handler()
        importer = Mock()
        importer.import_data.side_effect = ValueError
        FooImporter = Mock(return_value=importer)
        handler.importers = {'Foo': FooImporter}

        handler.commit_host_partial = False
        with patch.object(handler, 'commit_host_transaction') as commit:
            self.assertRaises(ValueError, handler.import_data, 'Foo')
            self.assertFalse(commit.called)

        handler.commit_host_partial = True
        with patch.object(handler, 'commit_host_transaction') as commit:
            self.assertRaises(ValueError, handler.import_data, 'Foo')
            commit.assert_called_once_with()


class BulkImportHandlerBattery(ImportHandlerBattery):

    def test_import_data_invalid_model(self):
        self.config = self.make_config()
        handler = self.make_handler()
        importer = Mock()
        importer.import_data.return_value = 0
        FooImporter = Mock(return_value=importer)
        handler.importers = {'Foo': FooImporter}

        handler.import_data('Foo')
        self.assertEqual(FooImporter.call_count, 1)
        importer.import_data.assert_called_once_with()

        FooImporter.reset_mock()
        importer.reset_mock()

        handler.import_data('Missing')
        self.assertFalse(FooImporter.called)
        self.assertFalse(importer.called)

    def test_import_data_with_changes(self):
        self.config = self.make_config()
        handler = self.make_handler()
        importer = Mock()
        FooImporter = Mock(return_value=importer)
        handler.importers = {'Foo': FooImporter}

        importer.import_data.return_value = 0
        with patch.object(handler, 'process_changes') as process:
            handler.import_data('Foo')
            self.assertFalse(process.called)

        importer.import_data.return_value = 3
        with patch.object(handler, 'process_changes') as process:
            handler.import_data('Foo')
            self.assertFalse(process.called)


class TestImportHandler(TestCase, ImportHandlerBattery):
    handler_class = handlers.ImportHandler

    def setUp(self):
        self.config = self.make_config()

    def test_batch_size_kwarg(self):
        with patch.object(handlers.ImportHandler, 'get_importers', Mock(return_value={'Foo': Importer})):

            # handler has no batch size by default
            handler = handlers.ImportHandler(self.config)
            self.assertFalse(hasattr(handler, 'batch_size'))

            # but can override with kwarg
            handler = handlers.ImportHandler(self.config, batch_size=42)
            self.assertEqual(handler.batch_size, 42)

            # importer default is 200
            handler = handlers.ImportHandler(self.config)
            importer = handler.get_importer('Foo')
            self.assertEqual(importer.batch_size, 200)

            # but can override with handler init kwarg
            handler = handlers.ImportHandler(self.config, batch_size=37)
            importer = handler.get_importer('Foo')
            self.assertEqual(importer.batch_size, 37)

            # can also override with get_importer kwarg
            handler = handlers.ImportHandler(self.config)
            importer = handler.get_importer('Foo', batch_size=98)
            self.assertEqual(importer.batch_size, 98)

    # TODO: must mock out AppHandler.send_email() instead
    # @patch('rattail.importing.handlers.send_email')
    # def test_process_changes_sends_email(self, send_email):
    #     handler = handlers.ImportHandler()
    #     handler.import_began = pytz.utc.localize(datetime.datetime.utcnow())
    #     changes = [], [], []

    #     # warnings disabled
    #     handler.warnings = False
    #     handler.process_changes(changes)
    #     self.assertFalse(send_email.called)

    #     # warnings enabled
    #     handler.warnings = True
    #     handler.process_changes(changes)
    #     self.assertEqual(send_email.call_count, 1)

    #     send_email.reset_mock()

    #     # warnings enabled, with command (just for coverage..)
    #     handler.warnings = True
    #     handler.command = Mock(name='import-testing', parent=Mock(name='rattail'))
    #     handler.process_changes(changes)
    #     self.assertEqual(send_email.call_count, 1)


class TestBulkImportHandler(TestCase, BulkImportHandlerBattery):
    handler_class = handlers.BulkImportHandler


######################################################################
# fake import handler, tested mostly for basic coverage
######################################################################

class MockImportHandler(handlers.ImportHandler):

    def get_importers(self):
        return {'Product': MockImporter}

    def import_data(self, *keys, **kwargs):
        result = super(MockImportHandler, self).import_data(*keys, **kwargs)
        self._result = result
        return result


class TestImportHandlerImportData(ImporterTester, TestCase):

    sample_data = OrderedDict([
        ('16oz', {'upc': '00074305001161', 'description': "Apple Cider Vinegar 16oz"}),
        ('32oz', {'upc': '00074305001321', 'description': "Apple Cider Vinegar 32oz"}),
        ('1gal', {'upc': '00074305011283', 'description': "Apple Cider Vinegar 1gal"}),
    ])

    def setUp(self):
        self.config = self.make_config()
        self.handler = MockImportHandler(config=self.config)
        self.importer = MockImporter(config=self.config)
        self.importer.handler = self.handler

    def make_config(self):
        config = make_config([], extend=False)
        config.setdefault('rattail', 'timezone.default', 'America/Chicago')
        return config

    def import_data(self, **kwargs):
        # must modify our importer in-place since we need the handler to return
        # that specific instance, below (because the host/local data context
        # managers reference that instance directly)
        self.importer._setup(**kwargs)
        with patch.object(self.handler, 'get_importer', Mock(return_value=self.importer)):
            result = self.handler.import_data('Product', **kwargs)
        if result:
            self.result = result['Product']
        else:
            self.result = [], [], []

    def test_invalid_importer_key_is_ignored(self):
        handler = handlers.ImportHandler(self.config)
        self.assertNotIn('InvalidKey', handler.importers)
        self.assertEqual(handler.import_data('InvalidKey'), {})

    def test_create(self):
        local = self.copy_data()
        del local['32oz']
        with self.host_data(self.sample_data):
            with self.local_data(local):
                self.import_data()
        self.assert_import_created('32oz')
        self.assert_import_updated()
        self.assert_import_deleted()

    def test_update(self):
        local = self.copy_data()
        local['16oz']['description'] = "wrong description"
        with self.host_data(self.sample_data):
            with self.local_data(local):
                self.import_data()
        self.assert_import_created()
        self.assert_import_updated('16oz')
        self.assert_import_deleted()

    def test_delete(self):
        local = self.copy_data()
        local['bogus'] = {'upc': '00000000000000', 'description': "Delete Me"}
        with self.host_data(self.sample_data):
            with self.local_data(local):
                self.import_data(delete=True)
        self.assert_import_created()
        self.assert_import_updated()
        self.assert_import_deleted('bogus')

    def test_duplicate(self):
        host = self.copy_data()
        host['32oz-dupe'] = host['32oz']
        with self.host_data(host):
            with self.local_data(self.sample_data):
                self.import_data()
        self.assert_import_created()
        self.assert_import_updated()
        self.assert_import_deleted()

    def test_max_create(self):
        local = self.copy_data()
        del local['16oz']
        del local['1gal']
        with self.host_data(self.sample_data):
            with self.local_data(local):
                self.import_data(max_create=1)
        self.assert_import_created('16oz')
        self.assert_import_updated()
        self.assert_import_deleted()

    def test_max_total_create(self):
        local = self.copy_data()
        del local['16oz']
        del local['1gal']
        with self.host_data(self.sample_data):
            with self.local_data(local):
                self.import_data(max_total=1)
        self.assert_import_created('16oz')
        self.assert_import_updated()
        self.assert_import_deleted()

    def test_max_update(self):
        local = self.copy_data()
        local['16oz']['description'] = "wrong"
        local['1gal']['description'] = "wrong"
        with self.host_data(self.sample_data):
            with self.local_data(local):
                self.import_data(max_update=1)
        self.assert_import_created()
        self.assert_import_updated('16oz')
        self.assert_import_deleted()

    def test_max_total_update(self):
        local = self.copy_data()
        local['16oz']['description'] = "wrong"
        local['1gal']['description'] = "wrong"
        with self.host_data(self.sample_data):
            with self.local_data(local):
                self.import_data(max_total=1)
        self.assert_import_created()
        self.assert_import_updated('16oz')
        self.assert_import_deleted()

    def test_max_delete(self):
        local = self.copy_data()
        local['bogus1'] = {'upc': '00000000000001', 'description': "Delete Me"}
        local['bogus2'] = {'upc': '00000000000002', 'description': "Delete Me"}
        with self.host_data(self.sample_data):
            with self.local_data(local):
                self.import_data(delete=True, max_delete=1)
        self.assert_import_created()
        self.assert_import_updated()
        self.assert_import_deleted('bogus1')

    def test_max_total_delete(self):
        local = self.copy_data()
        local['bogus1'] = {'upc': '00000000000001', 'description': "Delete Me"}
        local['bogus2'] = {'upc': '00000000000002', 'description': "Delete Me"}
        with self.host_data(self.sample_data):
            with self.local_data(local):
                self.import_data(delete=True, max_total=1)
        self.assert_import_created()
        self.assert_import_updated()
        self.assert_import_deleted('bogus1')

    def test_dry_run(self):
        local = self.copy_data()
        del local['32oz']
        local['16oz']['description'] = "wrong description"
        local['bogus'] = {'upc': '00000000000000', 'description': "Delete Me"}
        with self.host_data(self.sample_data):
            with self.local_data(local):
                self.import_data(delete=True, dry_run=True)
        # TODO: maybe need a way to confirm no changes actually made due to dry
        # run; currently results still reflect "proposed" changes.  this rather
        # bogus test is here just for coverage sake
        self.assert_import_created('32oz')
        self.assert_import_updated('16oz')
        self.assert_import_deleted('bogus')

    # TODO: must mock out AppHandler.send_email() instead
    # def test_warnings_run(self):
    #     local = self.copy_data()
    #     del local['32oz']
    #     local['16oz']['description'] = "wrong description"
    #     local['bogus'] = {'upc': '00000000000000', 'description': "Delete Me"}
    #     with self.host_data(self.sample_data):
    #         with self.local_data(local):
    #             with patch('rattail.importing.handlers.send_email') as send_email:
    #                 self.assertEqual(send_email.call_count, 0)
    #                 self.import_data(delete=True, warnings=True, dry_run=True)
    #                 self.assertEqual(send_email.call_count, 1)
    #     # second time is just for more coverage...
    #     with self.host_data(self.sample_data):
    #         with self.local_data(local):
    #             with patch('rattail.importing.handlers.send_email') as send_email:
    #                 self.handler.command = Mock()
    #                 self.assertEqual(send_email.call_count, 0)
    #                 self.import_data(delete=True, warnings=True)
    #                 self.assertEqual(send_email.call_count, 1)
    #     # TODO: maybe need a way to confirm no changes actually made due to dry
    #     # run; currently results still reflect "proposed" changes.  this rather
    #     # bogus test is here just for coverage sake
    #     self.assert_import_created('32oz')
    #     self.assert_import_updated('16oz')
    #     self.assert_import_deleted('bogus')


try:
    from sqlalchemy import orm
    from .test_postgresql import MockBulkImporter
except ImportError:
    pass
else:

    Session = orm.sessionmaker()


    class MockFromSQLAlchemyHandler(handlers.FromSQLAlchemyHandler):

        def make_host_session(self):
            return Session()


    class MockToSQLAlchemyHandler(handlers.ToSQLAlchemyHandler):

        def make_session(self):
            return Session()


    class TestFromSQLAlchemyHandler(TestCase):

        def test_init(self):
            handler = handlers.FromSQLAlchemyHandler()
            self.assertRaises(NotImplementedError, handler.make_host_session)

        def test_get_importer_kwargs(self):
            session = object()
            handler = handlers.FromSQLAlchemyHandler(host_session=session)
            kwargs = handler.get_importer_kwargs(None)
            self.assertEqual(list(kwargs.keys()), ['host_session'])
            self.assertIs(kwargs['host_session'], session)

        def test_begin_host_transaction(self):
            handler = MockFromSQLAlchemyHandler()
            self.assertIsNone(handler.host_session)
            handler.begin_host_transaction()
            self.assertIsInstance(handler.host_session, orm.Session)
            handler.host_session.close()

        def test_commit_host_transaction(self):
            # TODO: test actual commit for data changes
            session = Session()
            handler = handlers.FromSQLAlchemyHandler(host_session=session)
            self.assertIs(handler.host_session, session)
            handler.commit_host_transaction()
            self.assertIsNone(handler.host_session)

        def test_rollback_host_transaction(self):
            # TODO: test actual rollback for data changes
            session = Session()
            handler = handlers.FromSQLAlchemyHandler(host_session=session)
            self.assertIs(handler.host_session, session)
            handler.rollback_host_transaction()
            self.assertIsNone(handler.host_session)


    class TestToSQLAlchemyHandler(TestCase):

        def test_init(self):
            handler = handlers.ToSQLAlchemyHandler()
            self.assertRaises(NotImplementedError, handler.make_session)

        def test_get_importer_kwargs(self):
            session = object()
            handler = handlers.ToSQLAlchemyHandler(session=session)
            kwargs = handler.get_importer_kwargs(None)
            self.assertEqual(list(kwargs.keys()), ['session'])
            self.assertIs(kwargs['session'], session)

        def test_begin_local_transaction(self):
            handler = MockToSQLAlchemyHandler()
            self.assertIsNone(handler.session)
            handler.begin_local_transaction()
            self.assertIsInstance(handler.session, orm.Session)
            handler.session.close()

        def test_commit_local_transaction(self):
            # TODO: test actual commit for data changes
            session = Session()
            handler = handlers.ToSQLAlchemyHandler(session=session)
            self.assertIs(handler.session, session)
            with patch.object(handler, 'session') as session:
                handler.commit_local_transaction()
                session.commit.assert_called_once_with()
                self.assertFalse(session.rollback.called)
            # self.assertIsNone(handler.session)

        def test_rollback_local_transaction(self):
            # TODO: test actual rollback for data changes
            session = Session()
            handler = handlers.ToSQLAlchemyHandler(session=session)
            self.assertIs(handler.session, session)
            with patch.object(handler, 'session') as session:
                handler.rollback_local_transaction()
                session.rollback.assert_called_once_with()
                self.assertFalse(session.commit.called)
            # self.assertIsNone(handler.session)



# -*- coding: utf-8; -*-

from unittest import TestCase

from rattail.config import make_config

try:
    import sqlalchemy as sa
    from rattail import auth as mod
    from rattail.db import Session
except ImportError:
    pass
else:

    class TestAuthHandler(TestCase):

        def setUp(self):
            self.config = self.make_config()
            self.app = self.config.get_app()
            self.handler = self.make_handler()

        def make_config(self):
            return make_config([], extend=False)

        def make_handler(self):
            return mod.AuthHandler(self.config)

        def test_delete_user(self):
            engine = sa.create_engine('sqlite://')
            model = self.app.model
            model.Base.metadata.create_all(bind=engine)
            session = Session(bind=engine)

            # make a user, then delete - it should work
            user = model.User(username='foobar')
            session.add(user)
            session.commit()
            self.assertIn(user, session)
            self.handler.delete_user(user)
            session.commit()
            self.assertNotIn(user, session)

        def test_user_is_admin(self):
            engine = sa.create_engine('sqlite://')
            model = self.app.model
            model.Base.metadata.create_all(bind=engine)
            session = Session(bind=engine)

            # no user
            self.assertFalse(self.handler.user_is_admin(None))

            # real user but not an admin
            user = model.User(username='barney')
            session.add(user)
            session.commit()
            self.assertFalse(self.handler.user_is_admin(user))

            # but if they are admin, it shows
            admin = self.handler.get_role_administrator(session)
            user.roles.append(admin)
            session.commit()
            self.assertTrue(self.handler.user_is_admin(user))

            session.close()



# -*- coding: utf-8; -*-

from unittest import TestCase

from rattail import clientele
from rattail.config import make_config

try:
    import sqlalchemy as sa
except ImportError:
    pass
else:

    class TestClienteleHandler(TestCase):

        def setUp(self):
            self.config = make_config([], extend=False)
            self.app = self.config.get_app()
            self.handler = clientele.ClienteleHandler(self.config)
            self.engine = sa.create_engine('sqlite://')
            self.app.model.Base.metadata.create_all(bind=self.engine)

        def test_get_all_customers(self):
            model = self.app.model
            session = self.app.make_session(bind=self.engine)

            # add some customers
            session.add(model.Customer(name='Fred Flintstone'))
            session.add(model.Customer(name='Wilma Flintstone'))
            session.add(model.Customer(name='Barney Rubble', active_in_pos=False))

            # getting "all" will exclude inactive by default
            customers = self.handler.get_all_customers(session)
            self.assertEqual(len(customers), 2)

            # but we can request them too
            customers = self.handler.get_all_customers(session, include_inactive=True)
            self.assertEqual(len(customers), 3)

            session.commit()
            session.close()



# -*- coding: utf-8; -*-

try:
    from rattail.db import model
    from ... import DataTestCase
except ImportError:
    pass
else:

    class TestDataSyncChange(DataTestCase):

        def test_unicode(self):
            change = model.DataSyncChange()
            self.assertEqual(str(change), "(empty)")

            change = model.DataSyncChange(payload_type='Product', payload_key='00074305001321')
            self.assertEqual(str(change), "Product: 00074305001321")

            change = model.DataSyncChange(payload_type='Product', payload_key='00074305001321', deletion=True)
            self.assertEqual(str(change), "Product: 00074305001321 (deletion)")



# -*- coding: utf-8; -*-

import shutil
import tempfile
from unittest.mock import patch, Mock

from . import ImporterTester

try:
    from rattail.importing import rattail_bulk as bulk
    from .test_rattail import DualRattailTestCase
except ImportError:
    pass
else:

    class BulkImportTester(DualRattailTestCase, ImporterTester):

        handler_class = bulk.BulkFromRattailToRattail

        def setUp(self):
            self.setup_rattail()
            self.tempdir = tempfile.mkdtemp()
            self.config.setdefault('rattail', 'workdir', self.tempdir)
            self.config.setdefault('rattail', 'timezone.default', 'America/Chicago')
            self.handler = self.make_handler()

            # TODO: no-op for coverage, how lame is that
            self.handler.get_default_keys()

        def tearDown(self):
            self.teardown_rattail()
            shutil.rmtree(self.tempdir)

        @property
        def model_name(self):
            return self.make_importer().model_name

        def get_fields(self):
            return self.make_importer().fields

        def make_handler(self, **kwargs):
            if 'config' not in kwargs and hasattr(self, 'config'):
                kwargs['config'] = self.config
            return self.handler_class(**kwargs)

        def import_data(self, host_data=None, **kwargs):
            if host_data is None:
                fields = self.get_fields()
                host_data = list(self.copy_data().values())
                for data in host_data:
                    for field in fields:
                        data.setdefault(field, None)
            with patch.object(self.importer_class, 'normalize_host_data', Mock(return_value=host_data)):
                with patch.object(self.handler, 'make_host_session', Mock(return_value=self.host_session)):
                    return self.handler.import_data(self.model_name, **kwargs)


    class TestPersonImport(BulkImportTester):

        importer_class = bulk.PersonImporter

        sample_data = {
            'fred': {
                'uuid': 'fred',
                'first_name': 'Fred',
                'last_name': 'Flintstone',
            },
            'maurice': {
                'uuid': 'maurice',
                'first_name': 'Maurice',
                'last_name': 'Jones',
            },
            'zebra': {
                'uuid': 'zebra',
                'first_name': 'Zebra',
                'last_name': 'Jones',
            },
        }

        def test_create(self):
            if self.postgresql():
                result = self.import_data()
                self.assertEqual(result, {'Person': 3})

        def test_max_create(self):
            if self.postgresql():
                result = self.import_data(max_create=1)
                self.assertEqual(result, {'Person': 1})


    class TestProductImport(BulkImportTester):

        importer_class = bulk.ProductImporter

        def test_simple_fields(self):
            importer = self.make_importer()
            self.assertNotIn('regular_price_uuid', importer.simple_fields)
            self.assertNotIn('current_price_uuid', importer.simple_fields)



# -*- coding: utf-8; -*-

import os
import shutil
import tempfile
from unittest import TestCase

from rattail.config import make_config
from rattail.filemon import config_ as config
from rattail.filemon import Action
from rattail.exceptions import ConfigurationError


class TestProfile(TestCase):

    def setUp(self):
        self.config = make_config([], extend=False)
        self.config.setdefault('rattail.filemon', 'foo.actions', 'bar')

    def test_empty_config_means_empty_profile(self):
        profile = config.Profile(self.config, u'nonexistent_key')
        self.assertEqual(len(profile.dirs), 0)
        self.assertFalse(profile.watch_locks)
        self.assertTrue(profile.process_existing)
        self.assertFalse(profile.stop_on_error)
        self.assertEqual(len(profile.actions), 0)

    def test_action_must_specify_callable(self):
        self.assertRaises(ConfigurationError, config.Profile, self.config, u'foo')

    def test_action_must_not_specify_both_func_and_class_callables(self):
        self.config.setdefault('rattail.filemon', 'foo.action.bar.class', 'baz')
        self.config.setdefault('rattail.filemon', 'foo.action.bar.func', 'baz')
        self.assertRaises(ConfigurationError, config.Profile, self.config, u'foo')

    def test_action_with_func_callable(self):
        self.config.setdefault('rattail.filemon', 'foo.action.bar.func', 'os:remove')
        profile = config.Profile(self.config, u'foo')
        self.assertEqual(len(profile.actions), 1)
        action = profile.actions[0]
        self.assertEqual(action.spec, u'os:remove')
        self.assertTrue(action.action is os.remove)

    def test_action_with_class_callable(self):
        self.config.setdefault('rattail.filemon', 'foo.action.bar.class', 'rattail.filemon:Action')
        profile = config.Profile(self.config, u'foo')
        self.assertEqual(len(profile.actions), 1)
        action = profile.actions[0]
        self.assertEqual(action.spec, u'rattail.filemon:Action')
        self.assertTrue(isinstance(action.action, Action))

    def test_action_with_args(self):
        self.config.setdefault('rattail.filemon', 'foo.action.bar.func', 'shutil:move')
        self.config.setdefault('rattail.filemon', 'foo.action.bar.args', '/dev/null')
        profile = config.Profile(self.config, u'foo')
        self.assertEqual(len(profile.actions), 1)
        action = profile.actions[0]
        self.assertEqual(len(action.args), 1)
        self.assertEqual(action.args[0], u'/dev/null')

    def test_action_with_kwargs(self):
        self.config.setdefault('rattail.filemon', 'foo.action.bar.func', 'rattail.filemon.actions:raise_exception')
        self.config.setdefault('rattail.filemon', 'foo.action.bar.kwarg.message', "Hello World")
        profile = config.Profile(self.config, u'foo')
        self.assertEqual(len(profile.actions), 1)
        action = profile.actions[0]
        self.assertEqual(len(action.kwargs), 1)
        self.assertEqual(action.kwargs[u'message'], u"Hello World")

    def test_action_with_default_retry(self):
        self.config.setdefault('rattail.filemon', 'foo.action.bar.func', 'rattail.filemon.actions:noop')
        profile = config.Profile(self.config, u'foo')
        self.assertEqual(len(profile.actions), 1)
        action = profile.actions[0]
        self.assertEqual(action.retry_attempts, 1)
        self.assertEqual(action.retry_delay, 0)

    def test_action_with_valid_configured_retry(self):
        self.config.setdefault('rattail.filemon', 'foo.action.bar.func', 'rattail.filemon.actions:noop')
        self.config.setdefault('rattail.filemon', 'foo.action.bar.retry_attempts', '42')
        self.config.setdefault('rattail.filemon', 'foo.action.bar.retry_delay', '100')
        profile = config.Profile(self.config, u'foo')
        self.assertEqual(len(profile.actions), 1)
        action = profile.actions[0]
        self.assertEqual(action.retry_attempts, 42)
        self.assertEqual(action.retry_delay, 100)

    def test_action_with_invalid_configured_retry(self):
        self.config.setdefault('rattail.filemon', 'foo.action.bar.func', 'rattail.filemon.actions:noop')
        self.config.setdefault('rattail.filemon', 'foo.action.bar.retry_attempts', '-1')
        self.config.setdefault('rattail.filemon', 'foo.action.bar.retry_delay', '-1')
        profile = config.Profile(self.config, u'foo')
        self.assertEqual(len(profile.actions), 1)
        action = profile.actions[0]
        self.assertEqual(action.retry_attempts, 1)
        self.assertEqual(action.retry_delay, 0)

    def test_normalize_dirs(self):
        tempdir = tempfile.mkdtemp()
        dir1 = os.path.join(tempdir, 'dir1')
        os.makedirs(dir1)
        # dir2 will be pruned due to its not existing
        dir2 = os.path.join(tempdir, 'dir2')
        # file1 will be pruned due to its not being a directory
        file1 = os.path.join(tempdir, 'file1')
        with open(file1, 'wt') as f:
            f.write('')
        self.config.setdefault('rattail.filemon', 'foo.action.bar.func', 'os:remove')
        self.config.setdefault('rattail.filemon', 'foo.dirs', ' '.join(['"{0}"'.format(d) for d in [dir1, dir2, file1]]))
        profile = config.Profile(self.config, u'foo')
        self.assertEqual(len(profile.dirs), 1)
        self.assertEqual(profile.dirs[0], dir1)
        shutil.rmtree(tempdir)


class TestLoadProfiles(TestCase):

    def setUp(self):
        self.tempdir = tempfile.mkdtemp()
        self.config = self.make_config()

    def tearDown(self):
        shutil.rmtree(self.tempdir)

    def make_config(self, monitor=True, foo_dirs=True, foo_actions=True,
                    bar_dirs=True, bar_actions=True):
        cfg = make_config([], extend=False)
        if monitor:
            cfg.setdefault('rattail.filemon', 'monitor', 'foo, bar')
        if foo_dirs:
            cfg.setdefault('rattail.filemon', 'foo.dirs', '"{0}"'.format(self.tempdir))
        if foo_actions:
            cfg.setdefault('rattail.filemon', 'foo.actions', 'delete')
        cfg.setdefault('rattail.filemon', 'foo.action.delete.func', 'os:remove')
        if bar_dirs:
            cfg.setdefault('rattail.filemon', 'bar.dirs', '"{0}"'.format(self.tempdir))
        if bar_actions:
            cfg.setdefault('rattail.filemon', 'bar.actions', 'delete')
        cfg.setdefault('rattail.filemon', 'bar.action.delete.func', 'os:remove')
        return cfg

    def test_returns_all_profiles_specified_in_monitor_option(self):
        monitored = config.load_profiles(self.config)
        self.assertEqual(len(monitored), 2)

        # leave profiles intact but replace monitor option with one key only
        self.config = self.make_config(monitor=False)
        self.config.setdefault('rattail.filemon', 'monitor', 'foo')
        monitored = config.load_profiles(self.config)
        self.assertEqual(len(monitored), 1)

    def test_monitor_option_must_be_specified(self):
        self.config = self.make_config(monitor=False)
        self.assertRaises(ConfigurationError, config.load_profiles, self.config)

    def test_profiles_which_define_no_watched_folders_are_pruned(self):
        monitored = config.load_profiles(self.config)
        self.assertEqual(len(monitored), 2)
        # remove foo's watched folder(s)
        self.config = self.make_config(foo_dirs=False)
        monitored = config.load_profiles(self.config)
        self.assertEqual(len(monitored), 1)

    def test_profiles_which_define_no_actions_are_pruned(self):
        monitored = config.load_profiles(self.config)
        self.assertEqual(len(monitored), 2)
        # remove foo's actions
        self.config = self.make_config(foo_actions=False)
        monitored = config.load_profiles(self.config)
        self.assertEqual(len(monitored), 1)

    def test_fallback_to_legacy_mode(self):

        # replace 'monitor' option with 'monitored' and update profiles accordingly
        # TODO: This seems hacky.
        self.config = self.make_config(monitor=False, foo_dirs=False, foo_actions=False,
                                       bar_dirs=False, bar_actions=False)
        self.config.setdefault('rattail.filemon', 'monitored', 'foo,bar')

        self.config.setdefault('rattail.filemon', 'foo.dirs', "['{0}']".format(self.tempdir))
        self.config.setdefault('rattail.filemon', 'foo.actions', "['os:remove']")
        self.config.setdefault('rattail.filemon', 'bar.dirs', "['{0}']".format(self.tempdir))
        self.config.setdefault('rattail.filemon', 'bar.actions', "['os:remove']")

        # self.config.setdefault('rattail.filemon', 'foo.dirs', "['{0}']".format(self.tempdir))
        # self.config.setdefault('rattail.filemon', 'foo.actions', 'delete')
        # self.config.setdefault('rattail.filemon', 'foo.action.delete.func', "['os:remove']")
        # self.config.setdefault('rattail.filemon', 'bar.dirs', "['{0}']".format(self.tempdir))
        # self.config.setdefault('rattail.filemon', 'bar.actions', 'delete')
        # self.config.setdefault('rattail.filemon', 'bar.action.delete.func', "['os:remove']")

        monitored = config.load_profiles(self.config)
        self.assertEqual(len(monitored), 2)
        profiles = list(monitored.values())
        self.assertTrue(isinstance(profiles[0], config.LegacyProfile))
        self.assertTrue(isinstance(profiles[1], config.LegacyProfile))


class TestLegacyProfile(TestCase):

    def setUp(self):
        self.config = make_config([], extend=False)

    def test_empty_config_means_empty_profile(self):
        profile = config.LegacyProfile(self.config, u'nonexistent_key')
        self.assertEqual(len(profile.dirs), 0)
        self.assertFalse(profile.watch_locks)
        self.assertTrue(profile.process_existing)
        self.assertFalse(profile.stop_on_error)
        self.assertEqual(len(profile.actions), 0)

    def test_action_with_spec_only(self):
        self.config.setdefault('rattail.filemon', 'foo.actions', "['os:remove']")
        profile = config.LegacyProfile(self.config, u'foo')
        self.assertEqual(len(profile.actions), 1)
        spec, action, args, kwargs = profile.actions[0]
        self.assertEqual(spec, u'os:remove')
        self.assertTrue(action is os.remove)

    def test_action_with_spec_and_args(self):
        self.config.setdefault('rattail.filemon', 'foo.actions', "[('shutil:move', u'/dev/null')]")
        profile = config.LegacyProfile(self.config, u'foo')
        self.assertEqual(len(profile.actions), 1)
        spec, action, args, kwargs = profile.actions[0]
        self.assertEqual(spec, u'shutil:move')
        self.assertEqual(len(args), 1)
        self.assertEqual(args[0], u'/dev/null')

    def test_normalize_dirs(self):
        tempdir = tempfile.mkdtemp()
        dir1 = os.path.join(tempdir, 'dir1')
        os.makedirs(dir1)
        # dir2 will be pruned due to its not existing
        dir2 = os.path.join(tempdir, 'dir2')
        # file1 will be pruned due to its not being a directory
        file1 = os.path.join(tempdir, 'file1')
        with open(file1, 'wt') as f:
            f.write('')
        self.config.setdefault('rattail.filemon', 'foo.dirs',
                               "[{0}]".format(', '.join(["'{0}'".format(d) for d in [dir1, dir2, file1]])))
        profile = config.LegacyProfile(self.config, u'foo')
        self.assertEqual(len(profile.dirs), 1)
        self.assertEqual(profile.dirs[0], dir1)
        shutil.rmtree(tempdir)


class TestLoadLegacyProfiles(TestCase):

    def setUp(self):
        self.tempdir = tempfile.mkdtemp()
        self.config = self.make_config()

    def tearDown(self):
        shutil.rmtree(self.tempdir)

    def make_config(self, monitored=True, foo_dirs=True, foo_actions=True):
        cfg = make_config([], extend=False)
        if monitored:
            cfg.setdefault('rattail.filemon', 'monitored', 'foo, bar')
        if foo_dirs:
            cfg.setdefault('rattail.filemon', 'foo.dirs', "['{0}']".format(self.tempdir))
        if foo_actions:
            cfg.setdefault('rattail.filemon', 'foo.actions', "['os:remove']")
        cfg.setdefault('rattail.filemon', 'bar.dirs', "['{0}']".format(self.tempdir))
        cfg.setdefault('rattail.filemon', 'bar.actions', "['os:remove']")
        return cfg

    def test_returns_all_profiles_specified_in_monitor_option(self):

        monitored = config.load_legacy_profiles(self.config)
        self.assertEqual(len(monitored), 2)

        # leave profiles intact but replace monitored option with one key only
        self.config = self.make_config(monitored=False)
        self.config.setdefault('rattail.filemon', 'monitored', 'foo')
        monitored = config.load_legacy_profiles(self.config)
        self.assertEqual(len(monitored), 1)

    def test_monitor_option_must_be_specified(self):
        self.config = self.make_config(monitored=False)
        self.assertRaises(ConfigurationError, config.load_legacy_profiles, self.config)

    def test_profiles_which_define_no_watched_folders_are_pruned(self):
        monitored = config.load_legacy_profiles(self.config)
        self.assertEqual(len(monitored), 2)

        # remove foo's watched folder(s)
        self.config = self.make_config(foo_dirs=False)
        monitored = config.load_legacy_profiles(self.config)
        self.assertEqual(len(monitored), 1)

    def test_profiles_which_define_no_actions_are_pruned(self):
        monitored = config.load_legacy_profiles(self.config)
        self.assertEqual(len(monitored), 2)

        # remove foo's actions
        self.config = self.make_config(foo_actions=False)
        monitored = config.load_legacy_profiles(self.config)
        self.assertEqual(len(monitored), 1)



# -*- coding: utf-8; -*-

from unittest import TestCase

from rattail.config import make_config

try:
    import sqlalchemy as sa
    from rattail.autocomplete import departments as mod
    from rattail.db import Session
except ImportError:
    pass
else:

    class TestDepartmentAutocompleter(TestCase):

        def setUp(self):
            self.config = self.make_config()
            self.app = self.config.get_app()
            self.autocompleter = self.make_autocompleter()

        def make_config(self):
            return make_config([], extend=False)

        def make_autocompleter(self):
            return mod.DepartmentAutocompleter(self.config)

        def test_autocomplete(self):
            engine = sa.create_engine('sqlite://')
            model = self.app.model
            model.Base.metadata.create_all(bind=engine)
            session = Session(bind=engine)

            # first create a few departments
            grocery = model.Department(name='Grocery')
            session.add(grocery)
            wellness = model.Department(name='Wellness')
            session.add(wellness)
            bulk = model.Department(name='Bulk')
            session.add(bulk)

            # searching for nothing yields no results
            result = self.autocompleter.autocomplete(session, '')
            self.assertEqual(len(result), 0)

            # search for 'l' yields Wellness, Bulk
            result = self.autocompleter.autocomplete(session, 'l')
            self.assertEqual(len(result), 2)
            uuids = [info['value'] for info in result]
            self.assertIn(wellness.uuid, uuids)
            self.assertIn(bulk.uuid, uuids)

            # search for 'grocery' yields just that department
            result = self.autocompleter.autocomplete(session, 'grocery')
            self.assertEqual(len(result), 1)
            self.assertEqual(result[0]['value'], grocery.uuid)



# -*- coding: utf-8; -*-

from unittest import TestCase
from unittest.mock import Mock

try:
    from rattail.db import model
    from rattail.db.model import people
    from ... import DataTestCase
except ImportError:
    pass
else:

    class TestPerson(DataTestCase):

        def test_unicode(self):
            person = model.Person()
            self.assertEqual(str(person), "")

            person = model.Person(display_name="Fred Flintstone")
            self.assertEqual(str(person), "Fred Flintstone")

        # TODO: this is duplicated in TestPerson
        def test_add_email_address(self):
            person = model.Person()
            self.assertEqual(len(person.emails), 0)
            person.add_email_address('fred@mailinator.com')
            self.assertEqual(len(person.emails), 1)
            email = person.emails[0]
            self.assertEqual(email.type, 'Home')

            person = model.Person()
            self.assertEqual(len(person.emails), 0)
            person.add_email_address('fred@mailinator.com', type='Work')
            self.assertEqual(len(person.emails), 1)
            email = person.emails[0]
            self.assertEqual(email.type, 'Work')

        # TODO: this is duplicated in TestPerson
        def test_add_phone_number(self):
            person = model.Person()
            self.assertEqual(len(person.phones), 0)
            person.add_phone_number('417-555-1234')
            self.assertEqual(len(person.phones), 1)
            phone = person.phones[0]
            self.assertEqual(phone.type, 'Home')

            person = model.Person()
            self.assertEqual(len(person.phones), 0)
            person.add_phone_number('417-555-1234', type='Work')
            self.assertEqual(len(person.phones), 1)
            phone = person.phones[0]
            self.assertEqual(phone.type, 'Work')


    # TODO: deprecate/remove this?
    class TestFunctions(TestCase):

        def test_get_person_display_name(self):
            name = people.get_person_display_name("Fred", "Flintstone")
            self.assertEqual(name, "Fred Flintstone")

        def test_get_person_display_name_from_context(self):
            context = Mock(current_parameters={'first_name': "Fred", 'last_name': "Flintstone"})
            name = people.get_person_display_name_from_context(context)
            self.assertEqual(name, "Fred Flintstone")



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2024 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Typer-based command utilities
"""

import datetime
import inspect
import logging
from pathlib import Path
from typing import List, Optional

import makefun
import typer
from typing_extensions import Annotated
from wuttjamaican.util import load_entry_points

from rattail.config import make_config
from rattail.progress import ConsoleProgress, SocketProgress


def make_typer(**kwargs):
    """
    Create a Typer command instance, per Rattail conventions.

    This function is used to create the top-level ``rattail`` command,
    :data:`~rattail.commands.base.rattail_typer`.  You can use it to
    create additional top-level commands for your app, as needed.

    :returns: ``typer.Typer`` instance
    """
    kwargs.setdefault('callback', typer_callback)
    return typer.Typer(**kwargs)


def typer_eager_imports(
        group: [typer.Typer, str]):
    """
    Eagerly import all modules which are registered as having
    subcommands belonging to the given group.

    This is used to locate subcommands which may be defined by
    multiple different packages.  It is mostly needed for the main
    ``rattail`` command, since e.g. various POS integration packages
    may define additional subcommands for it.

    Most custom apps will define their own top-level command and some
    subcommands, but will have no need to "discover" additional
    subcommands defined elsewhere.  Hence you normally would not need
    to call this function.

    However if you wish to define a ``rattail`` subcommand(s), you
    *would* need to register the entry point for your module(s)
    containing the subcommand(s) like so (in ``pyproject.toml``):

    .. code-block:: ini

       [project.entry-points."rattail.typer_imports"]
       poser = "poser.commands"

    Note that ``rattail.typer_imports`` indicates you are registering
    a module which defines ``rattail`` subcommands.  The ``poser``
    name is arbitrary but should match your package name.

    :param group: Typer group command, or the name of one.
    """
    if isinstance(group, typer.Typer):
        group = group.info.name
    load_entry_points(f'{group}.typer_imports')


def typer_callback(
        ctx: typer.Context,

        ##############################

        # nb. these first args are defined in wuttjamaican

        config_paths: Annotated[
            Optional[List[Path]],
            typer.Option('--config', '-c',
                         exists=True,
                         help="Config path (may be specified more than once)")] = None,

        plus_config_paths: Annotated[
            Optional[List[Path]],
            typer.Option('--plus-config',
                         exists=True,
                         help="Extra configs to load in addition to normal config")] = None,

        progress: Annotated[
            bool,
            typer.Option('--progress', '-P',
                          help="Report progress when relevant")] = False,

        # # fn = click.option('-V', '--version',
        # #                   is_flag=True,
        # #                   help="Show program version"
        # #                   )(fn)

        # fn = click.option('--stdout',
        #                   type=click.File('wb'),
        #                   help="Optional file to which STDOUT should be written"
        #                   )(fn)

        # fn = click.option('--stderr',
        #                   type=click.File('wb'),
        #                   help="Optional file to which STDERR should be written"
        #                   )(fn)

        ##############################

        # nb. the rest of these args are defined in rattail

        no_init: Annotated[
            bool,
            typer.Option('--no-init', '-n')] = False,

        no_extend_config: Annotated[
            bool,
            typer.Option('--no-extend-config')] = False,

        verbose: Annotated[
            bool,
            typer.Option('--verbose')] = False,

        progress_socket: Annotated[
            str,
            typer.Option(help="Optional socket (e.g. localhost:8487) to which progress "
                         "info should be written.")] = None,

        runas_username: Annotated[
            str,
            typer.Option('--runas', '-R',
                         help="Optional username to impersonate when running the command.  "
                         "This is only relevant for / used by certain commands.")] = None,

        versioning: Annotated[
            bool,
            typer.Option('--versioning',
                         help="Force *enable* of data versioning.  If set, then --no-versioning "
                         "cannot also be set.  If neither is set, config will determine whether "
                         "or not data versioning should be enabled.")] = False,

        no_versioning: Annotated[
            bool,
            typer.Option('--no-versioning',
                         help="Force *disable* of data versioning.  If set, then --versioning "
                         "cannot also be set.  If neither is set, config will determine whether "
                         "or not data versioning should be enabled.")] = False,

):
    """
    Generic callback for use with top-level commands.
    """
    config = make_cli_config(ctx)

    ctx.rattail_config = config

    ctx.rattail_progress = None
    if progress:
        if progress_socket:
            host, port = progress_socket.split(':')
            ctx.rattail_progress = SocketProgress(host, int(port))
        else:
            ctx.rattail_progress = ConsoleProgress


def make_cli_config(ctx):
    """
    Make a config object according to the command-line context
    (params).

    :param ctx: ``typer.Context`` instance

    :returns: :class:`~rattail.config.RattailConfig` instance
    """
    kwargs = ctx.params

    # if args say not to "init" then we make a sort of empty config
    if kwargs.get('no_init'):
        config = make_config([], extend=False, versioning=False)

    else: # otherwise we make a proper config, and maybe turn on versioning
        logging.basicConfig()
        config = make_config(files=kwargs.get('config_paths') or None,
                             plus_files=kwargs.get('plus_config_paths') or None,
                             extend=not kwargs.get('no_extend_config'),
                             versioning=False)
        if kwargs.get('versioning'):
            from rattail.db.config import configure_versioning
            configure_versioning(config, force=True)
        elif not kwargs.get('no_versioning'):
            try:
                from rattail.db.config import configure_versioning
            except ImportError:
                pass
            else:
                configure_versioning(config)

    # import our primary data model now, just in case it hasn't fully been
    # imported yet.  this it to be sure association proxies and the like
    # are fully wired up in the case of extensions
    # TODO: what about web apps etc.? i guess i was only having the problem
    # for some importers, e.g. basic CORE API -> Rattail w/ the schema
    # extensions in place from rattail-corepos
    app = config.get_app()
    try:
        app.model
    except ImportError:
        pass

    return config


def typer_get_runas_user(ctx, session=None):
    """
    Convenience function to get the "runas" User object for the
    current command.

    Uses :meth:`rattail.app.AppHandler.get_runas_user()` under the
    hood, but the ``--runas`` command line param provides the default
    username.
    """
    config = ctx.parent.rattail_config
    app = config.get_app()
    return app.get_runas_user(session=session,
                              username=ctx.parent.params['runas_username'])


def importer_command_template(

        # model names (aka importer keys)
        models: Annotated[
            Optional[List[str]],
            typer.Argument(help="Which data models to import.  If you specify any, then only "
                           "data for those models will be imported.  If you do not specify "
                           "any, then all *default* models will be imported.")] = None,

        # list models
        list_all_models: Annotated[
            bool,
            typer.Option('--list-all-models', '-l',
                         help="List all available models and exit.")] = False,
        list_default_models: Annotated[
            bool,
            typer.Option('--list-default-models',
                         help="List the default models and exit.")] = False,

        # make batches
        make_batches: Annotated[
            bool,
            typer.Option('--make-batches',
                         help="If specified, make new Import / Export Batches instead of "
                         "performing an actual (possibly dry-run) import.")] = False,

        # key / fields / exclude
        key: Annotated[
            str,
            typer.Option('--key',
                         help="List of fields (comma-separated string) which should be "
                         "used as \"primary key\" for the import.")] = None,
        fields: Annotated[
            str,
            typer.Option('--fields',
                         help="List of fields (comma-separated string) which should be "
                         "included in the import.  If this parameter is specified, then "
                         "any field not listed here, would be *excluded* regardless of "
                         "the --exclude-fields parameter.")] = None,
        exclude_fields: Annotated[
            str,
            typer.Option('--exclude-fields',
                         help="List of fields (comma-separated string) which should be "
                         "excluded from the import.  Any field not listed here, would be "
                         "included (or not) depending on the --fields parameter and/or "
                         "default importer behavior.")] = None,

        # fuzzy fields
        fuzzy_fields: Annotated[
            str,
            typer.Option(help="List of fields (comma-separated string) for which diff "
                         "comparison should be \"fuzzy\".  This is intended for "
                         "timestamps and similar values which vary in granularity "
                         "between systems.")] = None,
        fuzz_factor: Annotated[
            int,
            typer.Option(help="Numeric value for use with --fuzzy-fields.  For "
                         "timestamp fields, this refers to the number of seconds "
                         "by which values are allowed to differ and still be "
                         "considered a match.")] = 1,

        # allow create?
        create: Annotated[
            bool,
            typer.Option(help="Allow new records to be created during the import.")] = True,
        max_create: Annotated[
            int,
            typer.Option(help="Maximum number of records which may be created, after which a "
                         "given import task should stop.  Note that this applies on a per-model "
                         "basis and not overall.")] = None,

        # allow update?
        update: Annotated[
            bool,
            typer.Option(help="Allow existing records to be updated during the import.")] = True,
        max_update: Annotated[
            int,
            typer.Option(help="Maximum number of records which may be updated, after which a "
                         "given import task should stop.  Note that this applies on a per-model "
                         "basis and not overall.")] = None,

        # allow delete?
        delete: Annotated[
            bool,
            typer.Option(help="Allow records to be deleted during the import.")] = False,
        max_delete: Annotated[
            int,
            typer.Option(help="Maximum number of records which may be deleted, after which a "
                         "given import task should stop.  Note that this applies on a per-model "
                         "basis and not overall.")] = None,

        # max total changes, per model
        max_total: Annotated[
            int,
            typer.Option(help="Maximum number of *any* record changes which may occur, after which "
                         "a given import task should stop.  Note that this applies on a per-model "
                         "basis and not overall.")] = None,

        # date ranges
        start_date: Annotated[
            datetime.datetime,
            typer.Option(formats=['%Y-%m-%d'],
                         help="Optional (inclusive) starting point for date range, by which host "
                         "data should be filtered.  Only used by certain importers.")] = None,
        end_date: Annotated[
            datetime.datetime,
            typer.Option(formats=['%Y-%m-%d'],
                         help="Optional (inclusive) ending point for date range, by which host "
                         "data should be filtered.  Only used by certain importers.")] = None,
        year: Annotated[
            int,
            typer.Option(help="Optional year, by which data should be filtered.  Only used "
                         "by certain importers.")] = None,

        # TODO: deprecate --batch, replace with --batch-size ?
        # batch size
        batch_size: Annotated[
            int,
            typer.Option('--batch',
                         help="Split work to be done into batches, with the specified number of "
                         "records in each batch.  Or, set this to 0 (zero) to disable batching. "
                         "Implementation for this may vary somewhat between importers.")] = 200,

        # collect changes for processing
        collect_changes: Annotated[
            bool,
            typer.Option(help="Collect changes for processing at the end of the run.  "
                         "This is required for warning diff emails, but disabling it "
                         "can help to cut down on memory usage during the run.")] = True,

        # email diff warnings
        warnings: Annotated[
            bool,
            typer.Option('--warnings', '-W',
                         help="Set this flag if you expect a \"clean\" import, and wish for any "
                         "changes which do occur to be processed further and/or specially.  The "
                         "behavior of this flag is ultimately up to the import handler, but the "
                         "default is to send an email notification.")] = False,
        max_diffs: Annotated[
            int,
            typer.Option(help="Maximum number of \"diffs\" to display per warning type, in a "
                         "warning email.  Only used if --warnings is in effect.")] = None,

        # dry run?
        dry_run: Annotated[
            bool,
            typer.Option('--dry-run',
                         help="Go through the full motions and allow logging etc. to "
                         "occur, but rollback (abort) the transaction at the end.  "
                         "Note that this flag is ignored if --make-batches is specified.")] = False,
):
    """
    Stub function which provides a common param signature; used with
    :func:`importer_command`.
    """


def importer_command(fn):
    """
    Decorator for import/export commands.  Adds common params based on
    :func:`importer_command_template`.
    """
    original_sig = inspect.signature(fn)
    reference_sig = inspect.signature(importer_command_template)

    params = list(original_sig.parameters.values())
    for i, param in enumerate(reference_sig.parameters.values()):
        params.insert(i + 1, param)

    # remove the **kwargs param
    params.pop(-1)

    final_sig = original_sig.replace(parameters=params)
    return makefun.create_function(final_sig, fn)


def file_exporter_command_template(
        output_dir: Annotated[
            Path,
            typer.Option(help="Directory to which output files should be written.  "
                         "Note that this is a *required* parameter.")] = ...,
):
    """
    Stub function to provide signature for exporter commands which
    produce data file(s) as output.  Used with
    :func:`file_exporter_command`.
    """


def file_exporter_command(fn):
    """
    Decorator for file export commands.  Adds common params based on
    :func:`file_exporter_command_template`.
    """
    original_sig = inspect.signature(fn)
    plain_import_sig = inspect.signature(importer_command_template)
    file_export_sig = inspect.signature(file_exporter_command_template)
    desired_params = (
        list(plain_import_sig.parameters.values())
        + list(file_export_sig.parameters.values()))

    params = list(original_sig.parameters.values())
    for i, param in enumerate(desired_params):
        params.insert(i + 1, param)

    # remove the **kwargs param
    params.pop(-1)

    final_sig = original_sig.replace(parameters=params)
    return makefun.create_function(final_sig, fn)


def file_importer_command_template(
        input_dir: Annotated[
            Path,
            typer.Option(exists=True,
                         help="Directory from which input files should be read.  "
                         "Note that this is a *required* parameter.")] = ...,
):
    """
    Stub function to provide signature for importer commands which
    require data file(s) as input.  Used with
    :func:`file_importer_command`.
    """


def file_importer_command(fn):
    """
    Decorator for file import commands.  Adds common params based on
    :func:`file_importer_command_template`.
    """
    original_sig = inspect.signature(fn)
    plain_import_sig = inspect.signature(importer_command_template)
    file_import_sig = inspect.signature(file_importer_command_template)
    desired_params = (
        list(plain_import_sig.parameters.values())
        + list(file_import_sig.parameters.values()))

    params = list(original_sig.parameters.values())
    for i, param in enumerate(desired_params):
        params.insert(i + 1, param)

    # remove the **kwargs param
    params.pop(-1)

    final_sig = original_sig.replace(parameters=params)
    return makefun.create_function(final_sig, fn)



# -*- coding: utf-8; -*-

from unittest import TestCase

import pytest

from rattail.vendors import handler as mod
from rattail.vendors.catalogs import CatalogParserNotFound
from rattail.config import make_config
from rattail.db import Session


class TestVendorHandler(TestCase):

    def setUp(self):
        self.config = self.make_config()
        self.app = self.config.get_app()
        self.handler = self.make_handler()

    def make_config(self):
        return make_config([], extend=False)

    def make_handler(self):
        return mod.VendorHandler(self.config)

    def test_choice_uses_dropdown(self):
        
        # do not use dropdown by default
        result = self.handler.choice_uses_dropdown()
        self.assertFalse(result)

        # but do use dropdown if so configured
        self.config.setdefault('rattail', 'vendors.choice_uses_dropdown',
                               'true')
        result = self.handler.choice_uses_dropdown()
        self.assertTrue(result)

    def test_get_vendor(self):
        try:
            import sqlalchemy as sa
        except ImportError:
            pytest.skip("test is not relevant without sqlalchemy")

        engine = sa.create_engine('sqlite://')
        model = self.app.model
        model.Base.metadata.create_all(bind=engine)
        session = Session(bind=engine)
        app = self.config.get_app()

        # no vendor if none exist yet!
        result = self.handler.get_vendor(session, 'acme')
        self.assertIsNone(result)

        # let's make the vendor and make sure uuid fetch works
        uuid = app.make_uuid()
        acme = model.Vendor(uuid=uuid, name="Acme")
        session.add(acme)
        result = self.handler.get_vendor(session, uuid)
        self.assertIs(result, acme)

        # if we search by key it still does not work
        result = self.handler.get_vendor(session, 'acme')
        self.assertIsNone(result)

        # but we can configure the key reference, then it will
        setting = model.Setting(name='rattail.vendor.acme', value=uuid)
        session.add(setting)
        result = self.handler.get_vendor(session, 'acme')
        self.assertIs(result, acme)

    def test_get_all_catalog_parsers(self):

        # some are always installed; make sure they come back
        Parsers = self.handler.get_all_catalog_parsers()
        self.assertTrue(len(Parsers))

    def test_get_supported_catalog_parsers(self):

        # by default all parsers are considered supported, so these
        # calls should effectively yield the same result
        all_parsers = self.handler.get_all_catalog_parsers()
        supported = self.handler.get_supported_catalog_parsers()
        self.assertEqual(len(all_parsers), len(supported))

        # now pretend only one is supported, using legacy setting
        self.config.setdefault('tailbone', 'batch.vendorcatalog.supported_parsers',
                               'rattail.contrib.generic')
        supported = self.handler.get_supported_catalog_parsers()
        self.assertEqual(len(supported), 1)
        Parser = supported[0]
        self.assertEqual(Parser.key, 'rattail.contrib.generic')

        # now pretend two are supported, using preferred setting
        self.config.setdefault('rattail', 'vendors.supported_catalog_parsers',
                               'rattail.contrib.generic, rattail.contrib.kehe')
        supported = self.handler.get_supported_catalog_parsers()
        self.assertEqual(len(supported), 2)
        keys = [Parser.key for Parser in supported]
        self.assertEqual(keys, ['rattail.contrib.generic', 'rattail.contrib.kehe'])

    def test_get_catalog_parser(self):
        
        # generic parser comes back fine
        parser = self.handler.get_catalog_parser('rattail.contrib.generic')
        self.assertIsNotNone(parser)
        self.assertEqual(parser.key, 'rattail.contrib.generic')

        # unknown key returns nothing
        parser = self.handler.get_catalog_parser('this_should_not_exist')
        self.assertIsNone(parser)

        # and can raise an error if we require
        self.assertRaises(CatalogParserNotFound, self.handler.get_catalog_parser,
                          'this_should_not_exist', require=True)



# -*- coding: utf-8; -*-

try:
    from rattail.db import model
    from ... import DataTestCase
except ImportError:
    pass
else:

    class TestUserEmailAddress(DataTestCase):

        def extra_setup(self):
            self.user = model.User(username='fred')
            self.session.add(self.user)
            self.session.flush()

        def test_email_defaults_to_none(self):
            self.assertTrue(self.user.get_email_address() is None)

        def test_email_comes_from_person_then_customer(self):
            # only customer has email at this point
            person = model.Person(first_name='Fred')
            customer = model.Customer(name='Fred')
            customer.add_email_address('customer@mailinator.com')
            customer.people.append(person)
            self.user.person = person
            self.session.add(customer)
            self.session.flush()
            self.assertEqual(self.user.get_email_address(), 'customer@mailinator.com')

            # now person email will take preference
            person.add_email_address('person@mailinator.com')
            self.session.refresh(person)
            self.assertEqual(self.user.get_email_address(), 'person@mailinator.com')

        def test_email_address_property_works_too(self):
            # even though this may go away some day, cover it for now
            person = model.Person(first_name='Fred')
            person.add_email_address('person@mailinator.com')
            self.user.person = person
            self.session.flush()
            self.assertEqual(self.user.email_address, 'person@mailinator.com')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2024 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Batch-related commands
"""

import datetime
import inspect
import json
import logging
import os
import sys
from pathlib import Path

import makefun
import typer
from typing_extensions import Annotated

from .base import rattail_typer
from .typer import typer_get_runas_user
from rattail.util import simple_error


log = logging.getLogger(__name__)


def batch_action_command_template(
        batch_type: Annotated[
            str,
            typer.Option(help="Type of batch to be dealt with, e.g. 'vendor_catalog'")] = ...,
        batch_uuid: Annotated[
            str,
            typer.Argument(help="UUID of the batch to be processed.")] = ...,
        dry_run: Annotated[
            bool,
            typer.Option('--dry-run',
                         help="Go through the full motions and allow logging etc. to "
                         "occur, but rollback (abort) the transaction at the end.")] = False,
):
    """
    Stub function which provides a common param signature; used with
    :func:`batch_action_command`.
    """


def batch_action_command(fn):
    """
    Decorator for batch handler commands.  Adds common params based on
    :func:`batch_action_command_template`.
    """
    original_sig = inspect.signature(fn)
    reference_sig = inspect.signature(batch_action_command_template)

    params = list(original_sig.parameters.values())
    for i, param in enumerate(reference_sig.parameters.values()):
        params.insert(i + 1, param)

    # remove the **kwargs param
    params.pop(-1)

    final_sig = original_sig.replace(parameters=params)
    return makefun.create_function(final_sig, fn)


@rattail_typer.command()
def make_batch(
        ctx: typer.Context,
        batch_type: Annotated[
            str,
            typer.Argument(help="Type of batch to be created, e.g. 'vendor_catalog'")] = ...,
        input_file: Annotated[
            Path,
            typer.Option(help="Path to single input file, to be used as data "
                         "source for the new batch.  (File format will vary "
                         "depending on batch type.)")] = ...,
        create_kwargs: Annotated[
            str,
            typer.Option('--kwargs',
                         parser=json.loads,
                         help="Optional JSON-encoded string containing extra "
                         "keyword arguments to be passed to the handler's batch "
                         "creation logic.")] = None,
        dry_run: Annotated[
            bool,
            typer.Option('--dry-run',
                         help="Go through the full motions and allow logging etc. to "
                         "occur, but rollback (abort) the transaction at the end.")] = False,
):
    """
    Make a new batch, from a data file
    """
    config = ctx.parent.rattail_config
    progress = ctx.parent.rattail_progress

    params = dict(ctx.params)

    user = typer_get_runas_user(ctx)
    do_run(config, params['batch_type'], do_make_batch, user, params, progress=progress)


@rattail_typer.command()
@batch_action_command
def auto_receive(
        ctx: typer.Context,
        **kwargs
):
    """
    Auto-receive all items in a receiving batch
    """
    config = ctx.parent.rattail_config
    progress = ctx.parent.rattail_progress

    params = dict(kwargs)
    params.update(ctx.params)

    user = typer_get_runas_user(ctx)
    do_run(config, params['batch_type'], do_auto_receive, user, params, progress=progress)


@rattail_typer.command()
@batch_action_command
def execute_batch(
        ctx: typer.Context,
        execute_kwargs: Annotated[
            str,
            typer.Option('--kwargs',
                         parser=json.loads,
                         help="Optional JSON-encoded string containing extra "
                         "keyword arguments to be passed to the handler's batch "
                         "execution function.")] = None,
        **kwargs
):
    """
    Execute a batch
    """
    config = ctx.parent.rattail_config
    progress = ctx.parent.rattail_progress

    params = dict(kwargs)
    params.update(ctx.params)

    user = typer_get_runas_user(ctx)
    do_run(config, params['batch_type'], do_execute_batch, user, params, progress=progress)


@rattail_typer.command()
@batch_action_command
def populate_batch(
        ctx: typer.Context,
        **kwargs
):
    """
    Populate initial data for a batch
    """
    config = ctx.parent.rattail_config
    progress = ctx.parent.rattail_progress

    params = dict(kwargs)
    params.update(ctx.params)

    user = typer_get_runas_user(ctx)
    do_run(config, params['batch_type'], do_populate_batch, user, params, progress=progress)


@rattail_typer.command()
@batch_action_command
def refresh_batch(
        ctx: typer.Context,
        **kwargs
):
    """
    Refresh data for a batch
    """
    config = ctx.parent.rattail_config
    progress = ctx.parent.rattail_progress

    params = dict(kwargs)
    params.update(ctx.params)

    user = typer_get_runas_user(ctx)
    do_run(config, params['batch_type'], do_refresh_batch, user, params, progress=progress)


@rattail_typer.command()
def purge_batches(
        ctx: typer.Context,
        list_types: Annotated[
            bool,
            typer.Option('--list-types',
                         help="If set, list available batch types instead of trying "
                         "to purge anything.")] = False,
        batch_type: Annotated[
            str,
            typer.Option(help="Type of batch to be purged, e.g. 'vendor_catalog'")] = ...,
        before: Annotated[
            datetime.datetime,
            typer.Option(formats=['%Y-%m-%d'],
                         help="Purge all batches executed prior to this date.  If not "
                         "specified, will use --before-days to calculate instead.")] = None,
        before_days: Annotated[
            int,
            typer.Option(formats=['%Y-%m-%d'],
                         help="Number of days before the current date, to be used "
                         "as the cutoff date if --before is not specified.  Default "
                         "is 90 days before current date.")] = 90,
        dry_run: Annotated[
            bool,
            typer.Option('--dry-run',
                         help="Go through the full motions and allow logging etc. to "
                         "occur, but rollback (abort) the transaction at the end.")] = False,
):
    """
    Purge old batches from the database
    """
    from rattail.batch.handlers import get_batch_types
    from rattail.db.util import finalize_session

    config = ctx.parent.rattail_config
    progress = ctx.parent.rattail_progress
    app = config.get_app()

    if list_types:
        keys = get_batch_types(config)
        for key in keys:
            sys.stdout.write(f"{key}\n")
        return

    handler = app.get_batch_handler(batch_type)
    session = app.make_session()

    kwargs = {'dry_run': dry_run,
              'progress': progress}
    if before:
        kwargs['before'] = app.localtime(before)
    else:
        kwargs['before_days'] = before_days

    handler.purge_batches(session, **kwargs)
    finalize_session(session, dry_run=dry_run)


def do_run(config, batch_type, action_func, user, params, progress=None):
    from rattail.db.util import finalize_session

    app = config.get_app()
    handler = app.get_batch_handler(batch_type)
    session = app.make_session()
    model = app.model
    user = session.get(model.User, user.uuid)

    try:
        success = action_func(config, params, handler, session, user, progress=progress)
    except Exception as error:
        log.warning("handler action failed", exc_info=True)
        # nb. only write "simple" error string to stdout, in case
        # caller (e.g. web app) is capturing that for display to
        # user.  admin can consult logs if more info is needed.
        # we do *not* write to stderr, because logging probably
        # is already doing that; we want to avoid mixing them.
        # TODO: this admittedly seems brittle, maybe caller should
        # pass command line flags to control this?
        sys.stdout.write(simple_error(error))
        session.rollback()
        session.close()
        sys.exit(42)

    finalize_session(session, dry_run=params['dry_run'], success=success)


def do_make_batch(config, params, handler, session, user, progress=None):
    """
    This will create a new batch of the specified type, then populate it
    with the given data file.
    """
    if not os.path.exists(params['input_file']):
        raise RuntimeError("input file path does not exist: {}".format(params['input_file']))

    app = config.get_app()
    kwargs = dict(params['create_kwargs'])
    delete_if_empty = kwargs.pop('delete_if_empty', False)
    auto_execute_allowed = kwargs.pop('auto_execute_allowed', True)

    batch = handler.make_batch(session, created_by=user, **kwargs)
    handler.set_input_file(batch, params['input_file'])
    handler.do_populate(batch, user)

    if delete_if_empty:
        session.flush()
        if not batch.data_rows:
            log.debug("auto-deleting empty '%s' batch: %s", handler.batch_key, batch)
            handler.do_delete(batch, dry_run=params['dry_run'],
                              progress=progress)
            batch = None

    if batch and auto_execute_allowed and handler.auto_executable(batch):
        handler.execute(batch, user=user)
        batch.executed = app.make_utc()
        batch.executed_by = user

    return True


def do_auto_receive(config, params, handler, session, user, progress=None):
    batch = session.get(handler.batch_model_class, params['batch_uuid'])
    if not batch:
        raise RuntimeError(f"Batch of type '{params['batch_type']}' "
                           "not found: {params['batch_uuid']}")

    return handler.auto_receive_all_items(batch, progress=progress)


def do_execute_batch(config, params, handler, session, user, progress=None):
    batch = session.get(handler.batch_model_class, params['batch_uuid'])
    if not batch:
        raise RuntimeError(f"Batch of type '{params['batch_type']}' "
                           "not found: {params['batch_uuid']}")

    return handler.do_execute(batch, user, progress=progress, **params['execute_kwargs'])


def do_populate_batch(config, params, handler, session, user, progress=None):
    batch = session.get(handler.batch_model_class, params['batch_uuid'])
    if not batch:
        raise RuntimeError(f"Batch of type '{params['batch_type']}' "
                           "not found: {params['batch_uuid']}")

    return handler.do_populate(batch, user, progress=progress)


def do_refresh_batch(config, params, handler, session, user, progress=None):
    batch = session.get(handler.batch_model_class, params['batch_uuid'])
    if not batch:
        raise RuntimeError(f"Batch of type '{params['batch_type']}' "
                           "not found: {params['batch_uuid']}")

    return handler.do_refresh(batch, user, progress=progress)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2024 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Importing Commands
"""

import datetime
import sys
import logging

import typer
from typing_extensions import Annotated

from rattail.app import GenericHandler
from .base import rattail_typer
from .typer import (typer_get_runas_user, importer_command,
                    file_importer_command, file_exporter_command)


log = logging.getLogger(__name__)


class ImportCommandHandler(GenericHandler):
    """
    Responsible for handling import/export command line runs.
    """

    def __init__(self, config, import_handler_key=None, import_handler_spec=None, **kwargs):
        super().__init__(config, **kwargs)

        if not import_handler_key and not import_handler_spec:
            raise ValueError("must provide either import_handler_key or import_handler_spec")

        self.import_handler_key = import_handler_key
        self.import_handler_spec = import_handler_spec

    def run(self, params, progress=None):

        if params['list_all_models']:
            self.list_all_models(params)
            return

        if params['list_default_models']:
            self.list_default_models(params)
            return

        handler = self.get_handler(params)
        models = params['models'] or handler.get_default_keys()
        log.debug("using handler: %s", handler)
        log.debug("importing models: %s", models)
        log.debug("params are: %s", params)

        kwargs = {
            'warnings': params['warnings'],
            'fields': self.config.parse_list(params['fields']),
            'exclude_fields': self.config.parse_list(params['exclude_fields']),
            'fuzzy_fields': self.config.parse_list(params['fuzzy_fields']),
            'fuzz_factor': params['fuzz_factor'],
            'create': params['create'],
            'max_create': params['max_create'],
            'update': params['update'],
            'max_update': params['max_update'],
            'delete': params['delete'],
            'max_delete': params['max_delete'],
            'max_total': params['max_total'],
            'start_date': params['start_date'],
            'end_date': params['end_date'],
            'progress': progress,
        }

        # ensure we have dates here, not datetime
        if kwargs['start_date'] and isinstance(kwargs['start_date'], datetime.datetime):
            kwargs['start_date'] = kwargs['start_date'].date()
        if kwargs['end_date'] and isinstance(kwargs['end_date'], datetime.datetime):
            kwargs['end_date'] = kwargs['end_date'].date()

        if params['make_batches']:
            kwargs.update({
                'runas_user': params['user'],
            })
            handler.make_batches(*models, **kwargs)
        else:
            kwargs.update({
                'key_fields': self.config.parse_list(params['key']) if params['key'] else None,
                'dry_run': params['dry_run'],
            })
            handler.import_data(*models, **kwargs)

        # TODO: should this logging happen elsewhere / be customizable?
        if params['dry_run']:
            log.info("dry run, so transaction was rolled back")
        else:
            log.info("transaction was committed")

    def get_handler_factory(self, params, **kwargs):
        """
        Should return an ImportHandler factory (e.g. class) which will
        later be called to create a handler instance.
        """
        # use explicit spec if one was provided
        if self.import_handler_spec:
            return self.app.load_object(self.import_handler_spec)

        # otherwise lookup the handler based on key
        # nb. normal logic returns an instance but we want its class
        handler = self.app.get_import_handler(self.import_handler_key, require=True)
        return type(handler)

    def get_handler(self, params, **kwargs):
        """
        Returns a handler instance to be used by the command.
        """
        factory = self.get_handler_factory(params)
        user = params['user']
        if user:
            kwargs.setdefault('runas_user', user)
            kwargs.setdefault('runas_username', user.username)
        kwargs.setdefault('dry_run', params['dry_run'])
        kwargs.setdefault('collect_changes_for_processing', params['collect_changes'])
        kwargs.setdefault('batch_size', params['batch_size'])
        if params['max_diffs']:
            kwargs.setdefault('diff_max_display', params['max_diffs'])
        if params.get('handler_kwargs'):
            kwargs.update(params['handler_kwargs'])
        return factory(self.config, **kwargs)

    def list_all_models(self, params):
        handler = self.get_handler(params)
        if not handler:
            sys.stderr.write("no handler configured!\n")
            if self.import_handler_key:
                sys.stderr.write(f"handler key is: {self.import_handler_key}\n")
            sys.exit(1)
        sys.stdout.write("ALL MODELS:\n")
        sys.stdout.write("==============================\n")
        defaults = handler.get_default_keys()
        for key in handler.get_importer_keys():
            sys.stdout.write(key)
            if key in defaults:
                sys.stdout.write(" (*)")
            sys.stdout.write("\n")
        sys.stdout.write("==============================\n")
        sys.stdout.write("(*) means also default\n")

    def list_default_models(self, params):
        handler = self.get_handler(params)
        sys.stdout.write("DEFAULT MODELS:\n")
        sys.stdout.write("==============================\n")
        for key in handler.get_default_keys():
            sys.stdout.write(f"{key}\n")


@rattail_typer.command()
@file_exporter_command
def export_csv(
        ctx: typer.Context,
        **kwargs
):
    """
    Export data from Rattail to CSV file(s)
    """
    config = ctx.parent.rattail_config
    progress = ctx.parent.rattail_progress
    handler = ImportCommandHandler(
        config, import_handler_key='to_csv.from_rattail.export')
    kwargs['user'] = typer_get_runas_user(ctx)
    kwargs['handler_kwargs'] = {'output_dir': kwargs['output_dir']}
    handler.run(kwargs, progress=progress)


@rattail_typer.command()
@importer_command
def export_rattail(
        ctx: typer.Context,
        dbkey: Annotated[
            str,
            typer.Option(help="Config key for database engine to be used as the \"target\" "
                         "Rattail system, i.e. where data will be exported.  This key must "
                         "be defined in the [rattail.db] section of your config file.")] = 'host',
        **kwargs
):
    """
    Export data to another Rattail database
    """
    config = ctx.parent.rattail_config
    progress = ctx.parent.rattail_progress
    handler = ImportCommandHandler(
        config, import_handler_key='to_rattail.from_rattail.export')
    kwargs['user'] = typer_get_runas_user(ctx)
    kwargs['handler_kwargs'] = {'dbkey': dbkey}
    handler.run(kwargs, progress=progress)


@rattail_typer.command()
@file_importer_command
def import_csv(
        ctx: typer.Context,
        **kwargs
):
    """
    Import data from CSV file(s) to Rattail database
    """
    config = ctx.parent.rattail_config
    progress = ctx.parent.rattail_progress
    handler = ImportCommandHandler(
        config, import_handler_key='to_rattail.from_csv.import')
    kwargs['user'] = typer_get_runas_user(ctx)
    kwargs['handler_kwargs'] = {'input_dir': kwargs['input_dir']}
    handler.run(kwargs, progress=progress)


@rattail_typer.command()
@file_importer_command
def import_ifps(
        ctx: typer.Context,
        **kwargs
):
    """
    Import data from IFPS file(s) to Rattail database
    """
    config = ctx.parent.rattail_config
    progress = ctx.parent.rattail_progress
    handler = ImportCommandHandler(
        config, import_handler_key='to_rattail.from_ifps.import')
    kwargs['user'] = typer_get_runas_user(ctx)
    kwargs['handler_kwargs'] = {'input_dir': kwargs['input_dir']}
    handler.run(kwargs, progress=progress)


@rattail_typer.command()
@importer_command
def import_rattail(
        ctx: typer.Context,
        dbkey: Annotated[
            str,
            typer.Option(help="Config key for database engine to be used as the Rattail "
                         "\"host\", i.e. the source of the data to be imported.  This key "
                         "must be defined in the [rattail.db] section of your config file.  "
                         "Defaults to 'host'.")] = 'host',
        **kwargs
):
    """
    Import data from another Rattail database
    """
    config = ctx.parent.rattail_config
    progress = ctx.parent.rattail_progress
    handler = ImportCommandHandler(
        config, import_handler_key='to_rattail.from_rattail.import')
    kwargs['user'] = typer_get_runas_user(ctx)
    kwargs['handler_kwargs'] = {'dbkey': dbkey}
    handler.run(kwargs, progress=progress)


@rattail_typer.command()
@importer_command
def import_versions(
        ctx: typer.Context,
        comment: Annotated[
            str,
            typer.Option('--comment', '-m',
                         help="Comment to be recorded with the transaction."
                         )] = "import catch-up versions",
        **kwargs
):
    """
    Make initial versioned records for data models
    """
    config = ctx.parent.rattail_config

    if not config.versioning_has_been_enabled:
        sys.stderr.write("Continuum versioning is not enabled, "
                         "per config and/or command line args\n")
        sys.exit(1)

    progress = ctx.parent.rattail_progress
    handler = ImportCommandHandler(
        config, import_handler_key='to_rattail_versions.from_rattail.import')
    kwargs['user'] = typer_get_runas_user(ctx)
    kwargs['handler_kwargs'] = {'comment': comment}
    handler.run(kwargs, progress=progress)
